/*
 * Tapioca XMPP Plugin
 * Copyright (c) 2005 INdT.
 * @author Andre Moreira Magalhaes <andre.magalhaes@indt.org.br> 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif 

#include "xmpp_audio_session.h"

#include "xmpp_call_c.h"
#include "xmpp_presence_push_c.h"
#include "xmpp_thread_c.h"

#include <tapioca/core/tapioca_enums.h>
#include <tapioca/core/tapioca_errors.h>
#include <tapioca/core/tapioca_utils.h>
#include <tapioca/core/tapioca_version.h>

#include <string.h>
#include <stdlib.h>

struct _XmppAudioSessionPrivate {
    xmpp_client_t client;
    xmpp_call_t call;
    const gchar *local_sdp;

    gboolean disposed;
};

static void     xmpp_audio_session_dispose                (GObject *object);

static gboolean xmpp_audio_session_add_members            (TapiocaSession *ts,
                                                           const gchar **uris,
                                                           GError **error);
static gboolean xmpp_audio_session_remove_members         (TapiocaSession *ts,
                                                           const gchar **uris,
                                                           GError **error);
static gboolean xmpp_audio_session_initiate               (TapiocaSession *ts,
                                                           GError **error);
static void     xmpp_audio_session_close                  (TapiocaSession *ts);
static gboolean xmpp_audio_session_set_local_capabilities (TapiocaAudioSession *tas,
                                                           const gchar *local_sdp,
                                                           GError **error);
static gboolean xmpp_audio_session_get_capabilities       (TapiocaAudioSession *tas,
                                                           gchar **local_sdp,
                                                           gchar **remote_sdp,
                                                           GError **error);
static gboolean xmpp_audio_session_add_local_candidate    (TapiocaAudioSession *tas, 
                                                           const gchar *name,
                                                           const gchar *protocol,
                                                           const gchar *ip,
                                                           guint port,
                                                           gdouble preference,
                                                           const gchar *username,
                                                           const gchar *password,
                                                           const gchar *type,
                                                           guint generation,
                                                           GError **error);
static gboolean xmpp_audio_session_accept                 (TapiocaAudioSession *tas,
                                                           GError **error);
static gboolean xmpp_audio_session_reject                 (TapiocaAudioSession *tas,
                                                           GError **error);

static void     add_codecs_to_session_description         (XmppAudioSession *xas,
                                                           xmpp_call_session_description_t session_description);
static gchar   *create_sdp_from_session_description       (XmppAudioSession *xas,
                                                           xmpp_call_session_description_t session_description);

static void     destroyed_cb                              (xmpp_call_t call, 
                                                           gpointer data);
static void     state_change_cb                           (xmpp_call_t call, 
                                                           gpointer data, 
                                                           gint state);
static void     error_cb                                  (xmpp_call_t call, 
                                                           gpointer data, 
                                                           gint error);
static void     remote_capabilities_updated_cb            (xmpp_call_t call, 
                                                           gpointer data, 
                                                           xmpp_call_session_description_t session_description);
static void     remote_candidate_added_cb                 (xmpp_call_t call, 
                                                           gpointer data, 
                                                           xmpp_candidate_t candidate);

G_DEFINE_TYPE(XmppAudioSession, xmpp_audio_session, TAPIOCA_TYPE_AUDIO_SESSION)

static void 
xmpp_audio_session_class_init (XmppAudioSessionClass *klass)
{
    GObjectClass *gobject_class;
    TapiocaSessionClass *tapioca_session_class;
    TapiocaAudioSessionClass *tapioca_audio_session_class;

    gobject_class = (GObjectClass *) klass;
    tapioca_session_class = (TapiocaSessionClass*) klass;
    tapioca_audio_session_class = (TapiocaAudioSessionClass*) klass;
    xmpp_audio_session_parent_class = g_type_class_peek_parent (klass);
    
    g_type_class_add_private (klass, sizeof (XmppAudioSessionPrivate));

    tapioca_session_class->add_members = xmpp_audio_session_add_members;
    tapioca_session_class->remove_members = xmpp_audio_session_remove_members;
    tapioca_session_class->initiate = xmpp_audio_session_initiate;
    tapioca_session_class->close = xmpp_audio_session_close;
    
    tapioca_audio_session_class->set_local_capabilities = xmpp_audio_session_set_local_capabilities;
    tapioca_audio_session_class->get_capabilities = xmpp_audio_session_get_capabilities;
    tapioca_audio_session_class->add_local_candidate = xmpp_audio_session_add_local_candidate;
    tapioca_audio_session_class->accept = xmpp_audio_session_accept;
    tapioca_audio_session_class->reject = xmpp_audio_session_reject;

    gobject_class->dispose = xmpp_audio_session_dispose;
}

static void 
xmpp_audio_session_init (XmppAudioSession *xas)
{
    xas->priv = XMPP_AUDIO_SESSION_GET_PRIVATE (xas);
    xas->priv->disposed = FALSE;
   
    xas->priv->client = NULL;
    xas->priv->call = NULL;
    xas->priv->local_sdp = NULL;
}

static void
xmpp_audio_session_dispose (GObject *object)
{
    XmppAudioSession *sc = XMPP_AUDIO_SESSION (object);

    if (sc->priv->disposed) {
        /* If dispose did already run, return. */
        return;
    }
    
    /* Make sure dispose does not run twice. */
    sc->priv->disposed = TRUE;
   
    G_OBJECT_CLASS (xmpp_audio_session_parent_class)->dispose (object);
}

XmppAudioSession *
xmpp_audio_session_new (xmpp_client_t client)
{
    XmppAudioSession *xas = XMPP_AUDIO_SESSION (g_object_new (XMPP_TYPE_AUDIO_SESSION, NULL));
    
    g_debug ("creating audio session...");
    xas->priv->client = client;
    g_debug ("session created...");
    
    return xas;
}

XmppAudioSession *
xmpp_audio_session_new_with_call (xmpp_client_t client,
                                  xmpp_call_t call)
{
    XmppAudioSession *xas = XMPP_AUDIO_SESSION (g_object_new (XMPP_TYPE_AUDIO_SESSION, NULL));
    
    g_debug ("creating session...");
    xas->priv->client = client;
    xas->priv->call = call;
    xmpp_call_set_destroyed_cb (call, destroyed_cb, xas);
    xmpp_call_set_state_change_cb (call, state_change_cb, xas);
    xmpp_call_set_error_cb (call, error_cb, xas);
    xmpp_call_set_remote_capabilities_updated_cb (call, remote_capabilities_updated_cb, xas);
    xmpp_call_set_remote_candidate_added_cb (call, remote_candidate_added_cb, xas);
    tapioca_session_change_state (TAPIOCA_SESSION (xas), TAPIOCA_SESSION_STATE_INITIATING);
    g_debug ("session created...");
    
    return xas;
}

static gboolean
xmpp_audio_session_add_members (TapiocaSession *ts,
                                const gchar **jids,
                                GError **error)
{
    guint i;
    if (g_list_length (ts->members) > 0) {
        for (i = 0; jids[i]; ++i) 
            tapioca_session_signal_add_member_error (ts, TAPIOCA_ERROR_INVALID_PARAMETERS, jids[i]);
        g_set_error (error, TAPIOCA_ERROR, TAPIOCA_ERROR_INVALID_PARAMETERS, 
                "Cannot add more than one member to this session");
        g_debug ("Cannot add more than one member to this session");
        return FALSE;
    }
   
    g_debug ("trying to add member %s", jids[0]);
    if (strchr (jids[0], '/')) {
        tapioca_session_signal_member_added (ts, jids[0]);
        for (i = 1; jids[i]; ++i) 
            tapioca_session_signal_add_member_error (ts, TAPIOCA_ERROR_INVALID_PARAMETERS, jids[i]);
    }
    else {
        tapioca_session_signal_add_member_error (ts, TAPIOCA_ERROR_INVALID_PARAMETERS, jids[0]);
        g_set_error (error, TAPIOCA_ERROR, TAPIOCA_ERROR_INVALID_PARAMETERS, 
                "Add member requires a full member jid at the form of node@domain/resource");
        g_debug ("Add member requires a full member jid at the form of node@domain/resource");
        return FALSE;
    }
    
    return TRUE;
}
 
static gboolean
xmpp_audio_session_remove_members (TapiocaSession *ts,
                                   const gchar **jids,
                                   GError **error)
{    
    // TODO check if user is in session before removing it
    tapioca_session_signal_member_removed (ts, jids[0]);
    return TRUE;
}
                                 
static gboolean
xmpp_audio_session_initiate (TapiocaSession *ts,
                             GError **error)
{
    XmppAudioSession *xas = XMPP_AUDIO_SESSION (ts);
    xmpp_presence_push_t presence_push;
    xmpp_call_session_description_t session_description;
 
    g_set_error_val_if_fail (xas->priv->call == NULL, FALSE, error, 
            TAPIOCA_ERROR, TAPIOCA_ERROR_GENERIC, "Cannot call initiate on an incoming call");
    
    presence_push = xmpp_client_get_presence_push (xas->priv->client);
    if (!xmpp_presence_push_has_phone_capability (presence_push, ts->members->data)) {
        g_debug ("user %s cannot receive calls", (const char *) ts->members->data);
        g_set_error (error, TAPIOCA_ERROR, TAPIOCA_ERROR_GENERIC, "User cannot receive calls");
        return FALSE;
    }
     
    g_debug ("initiating call");
    xas->priv->call = xmpp_call_session_client_create_call (
            xmpp_client_get_call_session_client (xas->priv->client));
    xmpp_call_set_destroyed_cb (xas->priv->call, destroyed_cb, xas);
    xmpp_call_set_state_change_cb (xas->priv->call, state_change_cb, xas);
    xmpp_call_set_error_cb (xas->priv->call, error_cb, xas);
    xmpp_call_set_remote_capabilities_updated_cb (xas->priv->call, remote_capabilities_updated_cb, xas);
    xmpp_call_set_remote_candidate_added_cb (xas->priv->call, remote_candidate_added_cb, xas);
    
    session_description = xmpp_call_session_description_new ();
    add_codecs_to_session_description (xas, session_description);
    
    // TODO set other cb
    xmpp_call_initiate (xas->priv->call, ts->members->data, session_description);

    return TRUE;
}

static void
xmpp_audio_session_close (TapiocaSession *ts)
{
    XmppAudioSession *xas = XMPP_AUDIO_SESSION (ts);
    
    xmpp_call_terminate (xas->priv->call);
}

static gboolean 
xmpp_audio_session_set_local_capabilities (TapiocaAudioSession *tas,
                                           const gchar *local_sdp,
                                           GError **error)
{
    XmppAudioSession *xas = XMPP_AUDIO_SESSION (tas);

    g_debug ("setting local capabilities");
    xas->priv->local_sdp = g_strdup (local_sdp);
    g_debug ("local capabilities set");

    return TRUE;
}

static gboolean 
xmpp_audio_session_add_local_candidate (TapiocaAudioSession *tas, 
                                        const gchar *name,
                                        const gchar *protocol,
                                        const gchar *ip,
                                        guint port,
                                        gdouble preference,
                                        const gchar *username,
                                        const gchar *password,
                                        const gchar *type,
                                        guint generation,
                                        GError **error)
{
    XmppAudioSession *xas = XMPP_AUDIO_SESSION (tas);
    xmpp_candidate_t candidate;
    
    candidate = xmpp_candidate_new ();
    xmpp_candidate_set_name (candidate, name);
    xmpp_candidate_set_protocol (candidate, protocol);
    xmpp_candidate_set_address (candidate, ip, port);
    xmpp_candidate_set_preference (candidate, preference);
    xmpp_candidate_set_username (candidate, username);
    xmpp_candidate_set_type (candidate, type);
    xmpp_candidate_set_generation (candidate, generation);
    
    xmpp_call_add_local_candidates (xas->priv->call, candidate, NULL);

    xmpp_candidate_destroy (candidate);

    return TRUE;
}

static gboolean 
xmpp_audio_session_get_capabilities (TapiocaAudioSession *tas,
                                     gchar **local_sdp,
                                     gchar **remote_sdp,
                                     GError **error)
{
    XmppAudioSession *xas = XMPP_AUDIO_SESSION (tas);
    
    g_debug ("getting capabilities...");
    // FIXME
    *local_sdp = g_strdup (xas->priv->local_sdp);
    *remote_sdp = create_sdp_from_session_description (xas, xmpp_call_get_remote_description (xas->priv->call));
    
    return TRUE;
}

static gboolean 
xmpp_audio_session_accept (TapiocaAudioSession *tas,
                           GError **error)
{
    XmppAudioSession *xas = XMPP_AUDIO_SESSION (tas);
    xmpp_call_session_description_t session_description;

    g_set_error_val_if_fail (xas->priv->call != NULL, FALSE, error, 
            TAPIOCA_ERROR, TAPIOCA_ERROR_GENERIC, "Cannot call accept on an outgoing call");
    g_set_error_val_if_fail (xas->priv->local_sdp != NULL, FALSE, error, 
            TAPIOCA_ERROR, TAPIOCA_ERROR_GENERIC, "No local capabilities set");
    
    g_debug ("accepting session...");
   
    session_description = xmpp_call_session_description_new ();
    add_codecs_to_session_description (xas, session_description);
    xmpp_call_accept (xas->priv->call, session_description);
    
    g_debug ("session accepted");

    return TRUE;
}

static gboolean 
xmpp_audio_session_reject (TapiocaAudioSession *tas,
                           GError **error)
{
    XmppAudioSession *xas = XMPP_AUDIO_SESSION (tas);

    g_set_error_val_if_fail (xas->priv->call != NULL, FALSE, error, 
            TAPIOCA_ERROR, TAPIOCA_ERROR_GENERIC, "Cannot call reject on an outgoing call");
    
    g_debug ("rejecting session...");
   
    xmpp_call_reject (xas->priv->call);
    
    g_debug ("session rejected");

    return TRUE;
}

static void
add_codecs_to_session_description (XmppAudioSession *xas,
                                   xmpp_call_session_description_t session_description)
{
    gchar *sdp;
    guint pt, pref = 1;
    gint rate, channels;
    gchar *encoding_name, *offset;
   
    // TODO check if the string is correctly formed

    // a=rtpmap:PT ENCODING_NAME/RATE[/CHANNELS]\r\n";
    sdp = strstr (xas->priv->local_sdp, "a=rtpmap:");
    while (sdp) {
        sdp = strchr (sdp, ':') + 1;    // sdp = "pt ENCODING_NAME/RATE[/CHANNELS]
        pt = strtoul (sdp, &sdp, 10);   // sdp = " ENCODING_NAME/RATE[/CHANNELS]
        ++sdp;                          // sdp = "ENCODING_NAME/RATE[/CHANNELS]
        offset = strchr (sdp, '/');
        encoding_name = g_strndup (sdp, offset - sdp);
        sdp = offset + 1;               // sdp = "RATE[/CHANNELS]
        rate = strtoul (sdp, &sdp, 10); // sdp = " [/CHANNELS]
        if (*sdp == '/') {
            ++sdp;                      // sdp = CHANNELS
            channels = strtoul (sdp, NULL, 10);
        }
        else
            channels = -1;

        if (channels != -1)
            g_debug ("adding codec %d %s/%d/%d", pt, encoding_name, rate, channels);
        else
            g_debug ("adding codec %d %s/%d", pt, encoding_name, rate);
        
        xmpp_call_session_description_add_codec (session_description, pt, encoding_name, rate, channels, pref++);

        g_free (encoding_name);

        sdp = strstr (sdp, "a=rtpmap:");
    }
}

static gchar *
create_sdp_from_session_description (XmppAudioSession *xas,
                                     xmpp_call_session_description_t session_description)
{
    guint i, count;
    gint pt, rate, channels, preference;
    const gchar *encoding_name = NULL;
    GString *sdp = g_string_new ("v=0\r\n"
            "o=- 123456 654321 IN IP4 127.0.0.1\r\n"
            "c=IN IP4 127.0.0.1\r\n");

    i = 0;
    count = xmpp_call_session_description_get_codecs_count (session_description);
    for (; i < count; ++i) {
        xmpp_call_session_description_get_codec (session_description, 
                i, &pt, &encoding_name, &rate, &channels, &preference);
        // TODO fix hardcoded rate when no rate is available
        g_string_append_printf (sdp, "a=rtpmap:%d %s/%d\r\n", 
                pt, encoding_name, (rate != -1 ? rate : 8000));                
    }
    
    return g_string_free (sdp, FALSE);
}

static void
destroyed_cb (xmpp_call_t call, 
              gpointer data)
{
    TapiocaSession *ts = TAPIOCA_SESSION (data);
    XmppAudioSession *xas = XMPP_AUDIO_SESSION (data);
    
    g_debug ("call destroyed");
    
    xas->priv->call = NULL;
    tapioca_session_change_state (ts, TAPIOCA_SESSION_STATE_CLOSED);
}

static void
state_change_cb (xmpp_call_t call, 
                 gpointer data, 
                 gint state)
{
    TapiocaSession *ts = TAPIOCA_SESSION (data);
    switch (state) {
        case 1:     // sent initiate
        /* case 2:     // received initiate */
            g_debug ("state changed to %d (initiate)", state);
            tapioca_session_change_state (ts, TAPIOCA_SESSION_STATE_INITIATING);
            break;
        case 8:     // received reject
        case 10:    // sent terminate
        case 11:    // received terminate
            g_debug ("state changed to %d (terminate)", state);
            break;
        case 12:    // in progress
            g_debug ("state changed to %d (in progress)", state);
            tapioca_session_change_state (ts, TAPIOCA_SESSION_STATE_ESTABLISHED);
            break;
        default:
            g_debug ("state changed to %d", state);
            break;
    }
}

static void 
error_cb (xmpp_call_t call, 
          gpointer data, 
          gint error)
{
    g_debug ("error on session with code %d", error);
}

static void 
remote_capabilities_updated_cb (xmpp_call_t call, 
                                gpointer data, 
                                xmpp_call_session_description_t session_description)
{
    XmppAudioSession *xas = XMPP_AUDIO_SESSION (data);
    gchar *remote_sdp;
    
    remote_sdp = create_sdp_from_session_description (xas, xmpp_call_get_remote_description (xas->priv->call));
    tapioca_audio_session_signal_remote_capabilities_updated (TAPIOCA_AUDIO_SESSION (xas), remote_sdp);
    g_free (remote_sdp);
}

static void 
remote_candidate_added_cb (xmpp_call_t call, 
                           gpointer data, 
                           xmpp_candidate_t candidate)
{
    const gchar *ip = NULL;
    guint port;
    
    xmpp_candidate_get_address (candidate, &ip, &port);

    g_debug ("remote candidate added: %s %s %s %d %f %s %s %s %d",
            xmpp_candidate_get_name (candidate),
            xmpp_candidate_get_protocol (candidate),
            ip,
            port,
            xmpp_candidate_get_preference (candidate),
            xmpp_candidate_get_username (candidate),
            xmpp_candidate_get_password (candidate),
            xmpp_candidate_get_type (candidate),
            xmpp_candidate_get_generation (candidate));

    tapioca_audio_session_signal_remote_candidate_added (
            TAPIOCA_AUDIO_SESSION (data),
            xmpp_candidate_get_name (candidate),
            xmpp_candidate_get_protocol (candidate),
            ip,
            port,
            xmpp_candidate_get_preference (candidate),
            xmpp_candidate_get_username (candidate),
            xmpp_candidate_get_password (candidate),
            xmpp_candidate_get_type (candidate),
            xmpp_candidate_get_generation (candidate));
}
