/*
 * Tapioca XMPP Plugin
 * Copyright (c) 2005 INdT.
 * @author Andre Moreira Magalhaes <andre.magalhaes@indt.org.br> 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "xmpp_call.h"
#include "xmpp_call_session_client.h"

#include <talk/p2p/base/sessionmessagefactory.h>

namespace {

const uint32 MSG_TIMEOUT = 1;
const uint32 MSG_ERROR = 2;
const uint32 MSG_STATE = 3;

}

XmppCall::XmppCall(XmppCallSessionClient *client, const std::string &name, const cricket::SessionID &id)
    : client_(client), 
      name_(name),
      id_(id),
      initiator_(false),
      description_(NULL),
      remote_description_(NULL),
      state_(cricket::Session::STATE_INIT),
      error_(cricket::Session::ERROR_NONE),
      destroyed_cb_(NULL),
      destroyed_cb_data_(NULL),
      state_change_cb_(NULL),
      state_change_cb_data_(NULL),
      error_cb_(NULL),
      error_cb_data_(NULL),
      remote_capabilities_updated_cb_(NULL),
      remote_capabilities_updated_cb_data_(NULL),
      remote_candidate_added_cb_(NULL),
      remote_candidate_added_cb_data_(NULL)
{
}

XmppCall::~XmppCall() 
{
    // TODO check this
    delete description_;
    delete remote_description_;
    
    if (destroyed_cb_) 
        destroyed_cb_((xmpp_call_t) this, destroyed_cb_data_);
}

bool XmppCall::Initiate(const std::string &to, const cricket::SessionDescription *description) 
{
    // Only from cricket::Session::STATE_INIT
    if (state_ != cricket::Session::STATE_INIT)
      return false;
    
    // Setup for signaling. Initiate is asynchronous. It occurs once the address
    // candidates are ready.
    initiator_ = true;
    remote_address_ = to;
    description_ = description;
    initiate_msgid_ = SendSessionMessage(cricket::SessionMessage::TYPE_INITIATE, description, NULL, NULL);
    SetState(cricket::Session::STATE_SENTINITIATE);
    
    cricket::ThreadManager::CurrentThread()->Clear(this, MSG_TIMEOUT);
    cricket::ThreadManager::CurrentThread()->PostDelayed(50 * 1000, this, MSG_TIMEOUT);
    return true;
}

bool XmppCall::Accept(const cricket::SessionDescription *description) 
{
    // Only if just received initiate
    if (state_ != cricket::Session::STATE_RECEIVEDINITIATE)
      return false;
    
    // Setup for signaling. Accept is asynchronous. It occurs once the address
    // candidates are ready.
    initiator_ = false;
    description_ = description;
    SendSessionMessage(cricket::SessionMessage::TYPE_ACCEPT, description, NULL, NULL);
    SetState(cricket::Session::STATE_SENTACCEPT);
    return true;
}

bool XmppCall::Modify(const cricket::SessionDescription *description) 
{
    // Only if session already cricket::Session::STATE_INPROGRESS
    if (state_ != cricket::Session::STATE_INPROGRESS)
        return false;
    
    // Modify is asynchronous. It occurs once the address candidates are ready.
    // Either side can send a modify. It is only valid in an already accepted
    // session.
    description_ = description;
    SendSessionMessage(cricket::SessionMessage::TYPE_MODIFY, description, NULL, NULL);
    SetState(cricket::Session::STATE_SENTMODIFY);
    
    // TODO start the session timeout
    cricket::ThreadManager::CurrentThread()->Clear(this, MSG_TIMEOUT);
    cricket::ThreadManager::CurrentThread()->PostDelayed(50 * 1000, this, MSG_TIMEOUT);
    return true;
}

bool XmppCall::Redirect(const std::string &target) 
{
    // Redirect is sent in response to an initiate or modify, to redirect the request
    if (state_ != cricket::Session::STATE_RECEIVEDINITIATE)
        return false;
    
    initiator_ = false;
    redirect_target_ = target;
    SendSessionMessage(cricket::SessionMessage::TYPE_REDIRECT, NULL, NULL, NULL);
    
    // A redirect puts us in the same state as reject.  It just sends a different
    // kind of reject message, if you like.
    SetState(cricket::Session::STATE_SENTREDIRECT);
    
    return true;
}

bool XmppCall::Reject() 
{
    // Reject is sent in response to an initiate or modify, to reject the request
    if (state_ != cricket::Session::STATE_RECEIVEDINITIATE && state_ != cricket::Session::STATE_RECEIVEDMODIFY)
        return false;
    
    initiator_ = false;
    SendSessionMessage(cricket::SessionMessage::TYPE_REJECT, NULL, NULL, NULL);
    SetState(cricket::Session::STATE_SENTREJECT);
    return true;
}

bool XmppCall::Terminate() 
{
    // Either side can terminate, at any time.
    if (state_ == cricket::Session::STATE_SENTTERMINATE && state_ != cricket::Session::STATE_RECEIVEDTERMINATE)
        return false;
    
    // But we don't need to terminate if we already rejected.  The other client
    // already knows that we're done with this session.
    if (state_ != cricket::Session::STATE_SENTREDIRECT)
        SendSessionMessage(cricket::SessionMessage::TYPE_TERMINATE, NULL, NULL, NULL);
    
    SetState(cricket::Session::STATE_SENTTERMINATE);
    return true;
}

void XmppCall::AddLocalCandidates(const std::vector<cricket::Candidate> &candidates) 
{
    SendSessionMessage(cricket::SessionMessage::TYPE_CANDIDATES, NULL, &candidates, NULL);
}

void XmppCall::SetSessionMessageFactory(cricket::SessionMessageFactory *factory) 
{
    session_message_factory_ = factory;
}

void XmppCall::SetState(cricket::Session::State state) 
{
    if (state != state_) {
        state_ = state;
        if (state_change_cb_) 
            state_change_cb_((xmpp_call_t) this, state_change_cb_data_, state);
        cricket::ThreadManager::CurrentThread()->Post(this, MSG_STATE);
    }
}

void XmppCall::SetError(cricket::Session::Error error) 
{
    if (error != error_) {
        error_ = error;
        if (error_cb_) 
            error_cb_((xmpp_call_t) this, error_cb_data_, error);
        cricket::ThreadManager::CurrentThread()->Post(this, MSG_ERROR);
    }
}

void XmppCall::OnMessage(cricket::Message *pmsg) 
{
    switch(pmsg->message_id) {
    case MSG_TIMEOUT:
        // Session timeout has occured. Check to see if the session is still trying
        // to signal. If so, the session has timed out.
        // The Sockets have their own timeout for connectivity.
        SetError(cricket::Session::ERROR_TIME);
        break;
    
    case MSG_ERROR:
        switch (error_) {
        case cricket::Session::ERROR_RESPONSE:
            // This state could be reached if we get an error in response to an IQ
            // or if the network is so slow we time out on an individual IQ exchange.
            // In either case, Terminate (send more messages) and ignore the likely
            // cascade of more errors.
    
            // fall through
        case cricket::Session::ERROR_NETWORK:
        case cricket::Session::ERROR_TIME:
            // Time ran out - no response
            Terminate();
            break;
    
        default:
            break;
      }
      break;
    
    case MSG_STATE:
        switch (state_) {
        case cricket::Session::STATE_SENTACCEPT:
        case cricket::Session::STATE_RECEIVEDACCEPT:        
            SetState(cricket::Session::STATE_INPROGRESS);
            cricket::ThreadManager::CurrentThread()->Clear(this, MSG_TIMEOUT);
            break;
   
        case cricket::Session::STATE_SENTREJECT:
        case cricket::Session::STATE_SENTREDIRECT:
        case cricket::Session::STATE_RECEIVEDREJECT:
            Terminate();
            break;
    
        case cricket::Session::STATE_SENTTERMINATE:
        case cricket::Session::STATE_RECEIVEDTERMINATE:
            client_->DestroyCall(this);
            break;
    
        default:
            // explicitly ignoring some states here
            break;
        }
        break;
    }
}

void XmppCall::OnIncomingMessage(const cricket::SessionMessage &m) 
{
    switch (m.type()) {
    case cricket::SessionMessage::TYPE_INITIATE:
        remote_description_ = m.description();
        if (remote_capabilities_updated_cb_)
            remote_capabilities_updated_cb_((xmpp_call_t) this, remote_capabilities_updated_cb_data_, (xmpp_call_session_description_t) remote_description_);
        remote_address_ = m.from();
        name_ = m.name();
        initiator_ = false;
        SetState(cricket::Session::STATE_RECEIVEDINITIATE);
        SendInitiateResponseMessage(&m);
        // if have candidates set send them
        break;
    
    case cricket::SessionMessage::TYPE_ACCEPT:
        remote_description_ = m.description();
        if (remote_capabilities_updated_cb_)
            remote_capabilities_updated_cb_((xmpp_call_t) this, remote_capabilities_updated_cb_data_, (xmpp_call_session_description_t) remote_description_);
        SendAcknowledgementMessage(&m);
        SetState(cricket::Session::STATE_RECEIVEDACCEPT);
        break;
    
    case cricket::SessionMessage::TYPE_MODIFY:
        remote_description_ = m.description();
        if (remote_capabilities_updated_cb_)
            remote_capabilities_updated_cb_((xmpp_call_t) this, remote_capabilities_updated_cb_data_, (xmpp_call_session_description_t) remote_description_);
        SendAcknowledgementMessage(&m);
        SetState(cricket::Session::STATE_RECEIVEDMODIFY);
        break;
    
    case cricket::SessionMessage::TYPE_CANDIDATES:
        if (remote_candidate_added_cb_) {
            for (std::vector<cricket::Candidate>::const_iterator it = m.candidates().begin(); it != m.candidates().end(); it++)
                remote_candidate_added_cb_((xmpp_call_t) this, remote_candidate_added_cb_data_, (xmpp_candidate_t) &(*it));
        }
        SendAcknowledgementMessage(&m);
        break;
    
    case cricket::SessionMessage::TYPE_REJECT:
        SendAcknowledgementMessage(&m);
        SetState(cricket::Session::STATE_RECEIVEDREJECT);
        break;
    
    case cricket::SessionMessage::TYPE_REDIRECT:
        OnRedirectMessage(m);
        SendAcknowledgementMessage(&m);
        break;
    
    case cricket::SessionMessage::TYPE_TERMINATE:
        SendAcknowledgementMessage(&m);
        SetState(cricket::Session::STATE_RECEIVEDTERMINATE);
        break;
    }
}

void XmppCall::OnIncomingResponse(const cricket::SessionResponseMessage &m) 
{
    if ((state_ == cricket::Session::STATE_SENTINITIATE || state_ == cricket::Session::STATE_RECEIVEDACCEPT) &&
        m.message_id() == initiate_msgid_) {
        // shhh TODO: Handle any response from the initiate as necessary
        initiate_msgid_.clear();
    }
}

void XmppCall::OnIncomingError(const cricket::SessionMessage &m) 
{
    // If a candidate message errors out or gets dropped for some reason we
    // ignore the error.
    if (m.type() != cricket::SessionMessage::TYPE_CANDIDATES) {
        SetError(cricket::Session::ERROR_RESPONSE);
    }
}

void XmppCall::OnRedirectMessage(const cricket::SessionMessage &m) 
{
    assert(state_ == cricket::Session::STATE_SENTINITIATE);
    if (state_ != cricket::Session::STATE_SENTINITIATE)
      return;
    
    assert(m.redirect_target().size() != 0);
    remote_address_ = m.redirect_target();
    
    SendSessionMessage(cricket::SessionMessage::TYPE_INITIATE, description_, NULL,
            m.redirect_cookie()->Copy());
    
    // Restart the session timeout.
    cricket::ThreadManager::CurrentThread()->Clear(this, MSG_TIMEOUT);
    cricket::ThreadManager::CurrentThread()->PostDelayed(50 * 1000, this, MSG_TIMEOUT);
    
    // resend candidates
}

std::string XmppCall::SendSessionMessage(cricket::SessionMessage::Type type,
                                         const cricket::SessionDescription *description,
                                         const std::vector<cricket::Candidate> *candidates,
                                         cricket::SessionMessage::Cookie *redirect_cookie) 
{
    assert(session_message_factory_ != NULL);
    cricket::SessionMessage *m = session_message_factory_->CreateSessionMessage();
    m->set_type(type);
    m->set_to(remote_address_);
    m->set_name(name_);
    m->set_description(description);
    m->set_session_id(id_);
    if (candidates)
      m->set_candidates(*candidates);
    m->set_redirect_target(redirect_target_);
    m->set_redirect_cookie(redirect_cookie);
    SignalOutgoingMessage(this, *m);
    
    std::string message_id = m->message_id();
    delete m;
    return message_id;
}

void XmppCall::SendInitiateResponseMessage(const cricket::SessionMessage *message) 
{
    cricket::SessionResponseMessage *m =
        message->CreateResponse(cricket::SessionResponseMessage::TYPE_INITIATERESPONSE);
    SignalSendResponse(this, m);
    delete m;
}

void XmppCall::SendAcknowledgementMessage(const cricket::SessionMessage *message) 
{
    cricket::SessionResponseMessage *m =
        message->CreateResponse(cricket::SessionResponseMessage::TYPE_OK);
    SignalSendResponse(this, m);
    delete m;
}
