/*
 * Copyright (c) 2005 INdT.
 * @author Andre Moreira Magalhaes <andre.magalhaes@indt.org.br> 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "xmpp_call_session_client.h"

#include "xmpp_call_session_description.h"
#include "xmpp_call_session_message_factory.h"
#include "xmpp_call_session_send_task.h"
#include "xmpp_call.h"

#include <talk/p2p/client/xmppsessionmessagefactory.h>
#include <talk/p2p/base/helpers.h>
#include <talk/p2p/base/sessionmanager.h>
#include <talk/base/logging.h>
#include <talk/xmllite/qname.h>
#include <talk/xmpp/constants.h>
#include <talk/xmllite/xmlelement.h>
#include <talk/xmllite/xmlprinter.h>

#include <iostream>

namespace {
    
// We only allow usernames to be this many characters or fewer.
const size_t kMaxUsernameSize = 16;

const std::string XMPP_CLIENT_CLASS_NAME("google-xmpp");

const std::string NS_GOOGLESESSION("http://www.google.com/session");
const std::string NS_PHONE("http://www.google.com/session/phone");
const std::string NS_EMPTY("");

const buzz::QName QN_GOOGLESESSION_SESSION(true, NS_GOOGLESESSION, "session");
const buzz::QName QN_GOOGLESESSION_CANDIDATE(true, NS_GOOGLESESSION, "candidate");
const buzz::QName QN_GOOGLESESSION_TARGET(true, NS_GOOGLESESSION, "target");
const buzz::QName QN_GOOGLESESSION_COOKIE(true, NS_GOOGLESESSION, "cookie");
const buzz::QName QN_GOOGLESESSION_REGARDING(true, NS_GOOGLESESSION, "regarding");

const buzz::QName QN_PHONE_DESCRIPTION(true, NS_PHONE, "description");
const buzz::QName QN_PHONE_PAYLOADTYPE(true, NS_PHONE, "payload-type");
const buzz::QName QN_PHONE_PAYLOADTYPE_ID(true, NS_EMPTY, "id");
const buzz::QName QN_PHONE_PAYLOADTYPE_NAME(true, NS_EMPTY, "name");
const buzz::QName QN_PHONE_PAYLOADTYPE_RATE(true, NS_EMPTY, "rate");
const buzz::QName QN_PHONE_PAYLOADTYPE_CHANNELS(true, NS_EMPTY, "channels");

const buzz::QName QN_ID(true, buzz::STR_EMPTY, "id");
const buzz::QName QN_INITIATOR(true, buzz::STR_EMPTY, "initiator");
const buzz::QName QN_NAME(true, buzz::STR_EMPTY, "name");
const buzz::QName QN_PORT(true, buzz::STR_EMPTY, "port");
const buzz::QName QN_NETWORK(true, buzz::STR_EMPTY, "network");
const buzz::QName QN_GENERATION(true, buzz::STR_EMPTY, "generation");
const buzz::QName QN_ADDRESS(true, buzz::STR_EMPTY, "address");
const buzz::QName QN_USERNAME(true, buzz::STR_EMPTY, "username");
const buzz::QName QN_PASSWORD(true, buzz::STR_EMPTY, "password");
const buzz::QName QN_PREFERENCE(true, buzz::STR_EMPTY, "preference");
const buzz::QName QN_PROTOCOL(true, buzz::STR_EMPTY, "protocol");
const buzz::QName QN_KEY(true, buzz::STR_EMPTY, "key");

}

class XmlCookie: public cricket::SessionMessage::Cookie {
public:
    XmlCookie(const buzz::XmlElement* elem) : elem_(new buzz::XmlElement(*elem)) { }
    virtual ~XmlCookie() { delete elem_; }
    const buzz::XmlElement* elem() const { return elem_; }
    virtual Cookie* Copy() { return new XmlCookie(elem_); }
private:
    buzz::XmlElement* elem_;
};

XmppCallSessionClient::XmppCallSessionClient(buzz::Task *parent) 
    : buzz::XmppTask(parent, buzz::XmppEngine::HL_SINGLE) 
{    
    session_message_factory_ = new XmppCallSessionMessageFactory();
}

XmppCallSessionClient::~XmppCallSessionClient()
{
    delete session_message_factory_;
}

bool XmppCallSessionClient::HandleStanza(const buzz::XmlElement *stanza) 
{
    // Is it a IQ set stanza?
    if (stanza->Name() != buzz::QN_IQ)
        return false;
    if (stanza->Attr(buzz::QN_TYPE) != buzz::STR_SET)
        return false;
    
    // Make sure it has the right child element
    const buzz::XmlElement* element
        = stanza->FirstNamed(QN_GOOGLESESSION_SESSION);
    if (element == NULL)
        return false;
    
    // Is it one of the allowed types?
    std::string type;
    if (element->HasAttr(buzz::QN_TYPE)) {
        type = element->Attr(buzz::QN_TYPE);
        
        if (type != "initiate" && type != "accept" && type != "modify" &&
            type != "candidates" && type != "reject" && type != "redirect" &&
            type != "terminate") 
            return false;
    }
    
    // Does this client own the session description namespace?
    buzz::QName qn_session_desc(NS_PHONE, "description");
    const buzz::XmlElement* description = element->FirstNamed(qn_session_desc);
    if (type == "initiate" || type == "accept" || type == "modify") {
        if (description == NULL)
            return false;
    } else {
        if (description != NULL)
            return false;
    }

    QueueStanza(stanza);
    return true;
}

XmppCall *XmppCallSessionClient::CreateCall() 
{
    return CreateCall(NS_PHONE, cricket::SessionID(GetClient()->jid().Str(), cricket::CreateRandomId()), false);
}

XmppCall *XmppCallSessionClient::CreateCall(const std::string &name, 
                                            const cricket::SessionID &id, 
                                            bool received_initiate) 
{
    cricket::CritScope s(&crit_);
    XmppCall *call = new XmppCall(this, name, id);
    call->SetSessionMessageFactory(session_message_factory_);
    call->SignalSendResponse.connect(this, &XmppCallSessionClient::OnSendResponse);
    call->SignalOutgoingMessage.connect(this, &XmppCallSessionClient::OnOutgoingMessage);
    calls_[call->id()] = call;
    SignalCallCreate(call, received_initiate);
    return call;
}

void XmppCallSessionClient::DestroyCall(XmppCall *call)
{
    if (call == NULL)
        return;

    cricket::CritScope s(&crit_);
    std::map<cricket::SessionID, XmppCall *>::iterator it = calls_.find(call->id());
    if (it != calls_.end()) {
        SignalCallDestroy(call);
        calls_.erase(it);
        delete call;
    }
}

void XmppCallSessionClient::TerminateAll() 
{
    while (calls_.begin() != calls_.end()) {
        XmppCall *call = calls_.begin()->second;
        call->Terminate();
    }
}

XmppCall *XmppCallSessionClient::GetCall(const cricket::SessionID& id) 
{
    std::map<cricket::SessionID, XmppCall *>::iterator it = calls_.find(id);
    if (it != calls_.end())
        return it->second;
    return NULL;
}

int XmppCallSessionClient::ProcessStart() 
{
    const buzz::XmlElement *stanza = NextStanza();
    if (stanza == NULL)
        return STATE_BLOCKED;

    cricket::XmppSessionMessage message;
    ParseIncomingMessage(stanza, message);

    HandleIncomingMessage(message);
    
    return STATE_START;
}

bool XmppCallSessionClient::ParseIncomingMessage(const buzz::XmlElement *stanza,
                                                 cricket::XmppSessionMessage& message) 
{
    // Parse stanza into cricket::SessionMessage
    const buzz::XmlElement* element = stanza->FirstNamed(QN_GOOGLESESSION_SESSION);
    
    if (stanza->HasAttr(QN_ID))
        message.set_message_id(stanza->Attr(QN_ID));
    message.set_stanza(stanza);
    
    std::string type = element->Attr(buzz::QN_TYPE);
    if (type == "initiate" || type == "accept" || type == "modify") {
        ParseInitiateAcceptModify(stanza, message);
    } else if (type == "candidates") {
        ParseCandidates(stanza, message);
    } else if (type == "reject" || type == "terminate") {
        ParseRejectTerminate(stanza, message);
    } else if (type == "redirect") {
        ParseRedirect(stanza, message);
    } else {
        return false;
    }
    
    return true;
}

bool XmppCallSessionClient::ParseIncomingResponse(const buzz::XmlElement *stanza,
                                                  cricket::XmppSessionResponseMessage& response) 
{
    assert(stanza->Attr(buzz::QN_TYPE) == buzz::STR_RESULT);
    response.set_message_id(stanza->Attr(QN_ID));
    response.set_type(cricket::SessionResponseMessage::TYPE_OK);
    return true;
}

void XmppCallSessionClient::ParseHeader(const buzz::XmlElement *stanza, 
                                        cricket::XmppSessionMessage &message) 
{
    if (stanza->HasAttr(buzz::QN_FROM))
        message.set_from(stanza->Attr(buzz::QN_FROM));
    if (stanza->HasAttr(buzz::QN_TO))
        message.set_to(stanza->Attr(buzz::QN_TO));
    
    const buzz::XmlElement *element = stanza->FirstNamed(QN_GOOGLESESSION_SESSION);
    if (element->HasAttr(QN_ID))
        message.session_id().set_id_str(element->Attr(QN_ID));
    
    if (element->HasAttr(QN_INITIATOR))
        message.session_id().set_initiator(element->Attr(QN_INITIATOR));
    
    std::string type = element->Attr(buzz::QN_TYPE);
    if (type == "initiate") {
        message.set_type(cricket::SessionMessage::TYPE_INITIATE);
    } else if (type == "accept") {
        message.set_type(cricket::SessionMessage::TYPE_ACCEPT);
    } else if (type == "modify") {
        message.set_type(cricket::SessionMessage::TYPE_MODIFY);
    } else if (type == "candidates") {
        message.set_type(cricket::SessionMessage::TYPE_CANDIDATES);
    } else if (type == "reject") {
        message.set_type(cricket::SessionMessage::TYPE_REJECT);
    } else if (type == "redirect") {
        message.set_type(cricket::SessionMessage::TYPE_REDIRECT);
    } else if (type == "terminate") {
        message.set_type(cricket::SessionMessage::TYPE_TERMINATE);
    } else {
        assert(false);
    }
}

void XmppCallSessionClient::ParseInitiateAcceptModify(const buzz::XmlElement *stanza,
                                                      cricket::XmppSessionMessage &message) 
{
    // Pull the standard header pieces out
    ParseHeader(stanza, message);
    
    // Parse session description
    const buzz::XmlElement *session = stanza->FirstNamed(QN_GOOGLESESSION_SESSION);
    assert(session != NULL);
    buzz::QName qn_session_desc(NS_PHONE, "description");
    const buzz::XmlElement* desc_elem = session->FirstNamed(qn_session_desc);
    const cricket::SessionDescription *description = NULL;
    if (desc_elem)
        description = CreateSessionDescription(desc_elem);
    message.set_name(NS_PHONE);
    message.set_description(description);
}

void XmppCallSessionClient::ParseCandidates(const buzz::XmlElement *stanza, 
                                            cricket::XmppSessionMessage &message) 
{
    // Pull the standard header pieces out
    ParseHeader(stanza, message);
    
    // Parse candidates and session description
    std::vector<cricket::Candidate> candidates;
    const buzz::XmlElement *element = stanza->FirstNamed(QN_GOOGLESESSION_SESSION);
    const buzz::XmlElement *child = element->FirstElement();
    while (child != NULL) {
        if (child->Name() == QN_GOOGLESESSION_CANDIDATE) {
            cricket::Candidate candidate;
            if (ParseCandidate(child, &candidate))
                candidates.push_back(candidate);
        }
        child = child->NextElement();
    }
    message.set_name(NS_PHONE);
    message.set_candidates(candidates);
}

bool XmppCallSessionClient::ParseCandidate(const buzz::XmlElement *child,
                                           cricket::Candidate *candidate) 
{
    // Check for all of the required attributes.
    if (!child->HasAttr(QN_NAME) ||
        !child->HasAttr(QN_ADDRESS) ||
        !child->HasAttr(QN_PORT) ||
        !child->HasAttr(QN_USERNAME) ||
        !child->HasAttr(QN_PREFERENCE) ||
        !child->HasAttr(QN_PROTOCOL) ||
        !child->HasAttr(QN_GENERATION)) {
        LOG(LERROR) << "Candidate missing required attribute";
        return false;
    }
    
    cricket::SocketAddress address;
    address.SetIP(child->Attr(QN_ADDRESS));
    std::istringstream ist(child->Attr(QN_PORT));
    int port;
    ist >> port;
    address.SetPort(port);
    
    if (address.IsAny()) {
        LOG(LERROR) << "Candidate has address 0";
        return false;
    }
    
    // Always disallow addresses that refer to the local host.
    if (address.IsLocalIP()) {
        LOG(LERROR) << "Candidate has local IP address";
        return false;
    }
    
    // Disallow all ports below 1024, except for 80 and 443 on public addresses.
    if (port < 1024) {
        if ((port != 80) && (port != 443)) {
            LOG(LERROR) << "Candidate has port below 1024, not 80 or 443";
            return false;
        }
        if (address.IsPrivateIP()) {
            LOG(LERROR) << "Candidate has port of 80 or 443 with private IP address";
            return false;
        }
    }
    
    candidate->set_name(child->Attr(QN_NAME));
    candidate->set_address(address);
    candidate->set_username(child->Attr(QN_USERNAME));
    candidate->set_preference_str(child->Attr(QN_PREFERENCE));
    candidate->set_protocol(child->Attr(QN_PROTOCOL));
    candidate->set_generation_str(child->Attr(QN_GENERATION));
    
    // Check that the username is not too long and does not use any bad chars.
    if (candidate->username().size() > kMaxUsernameSize) {
        LOG(LERROR) << "Candidate username is too long";
        return false;
    }
    if (!cricket::IsBase64Encoded(candidate->username())) {
        LOG(LERROR) << "Candidate username has non-base64 encoded characters";
        return false;
    }
    
    // Look for the non-required attributes.
    if (child->HasAttr(QN_PASSWORD))
        candidate->set_password(child->Attr(QN_PASSWORD));
    if (child->HasAttr(buzz::QN_TYPE))
        candidate->set_type(child->Attr(buzz::QN_TYPE));
    if (child->HasAttr(QN_NETWORK))
        candidate->set_network_name(child->Attr(QN_NETWORK));
    
    return true;
}

void XmppCallSessionClient::ParseRejectTerminate(const buzz::XmlElement *stanza, 
                                                 cricket::XmppSessionMessage &message) 
{
    // Reject and terminate are very simple
    ParseHeader(stanza, message);
}

void XmppCallSessionClient::ParseRedirect(const buzz::XmlElement *stanza, 
                                          cricket::XmppSessionMessage &message) 
{
    // Pull the standard header pieces out
    ParseHeader(stanza, message);
    const buzz::XmlElement *session = stanza->FirstNamed(QN_GOOGLESESSION_SESSION);
    
    // Parse the target and cookie.
    const buzz::XmlElement* target = session->FirstNamed(QN_GOOGLESESSION_TARGET);
    if (target)
        message.set_redirect_target(target->Attr(QN_NAME));
    
    const buzz::XmlElement* cookie = session->FirstNamed(QN_GOOGLESESSION_COOKIE);
    if (cookie)
        message.set_redirect_cookie(new XmlCookie(cookie));
}

buzz::XmlElement *XmppCallSessionClient::TranslateHeader(const cricket::XmppSessionMessage &message) 
{
    buzz::XmlElement *result = new buzz::XmlElement(buzz::QN_IQ);
    result->AddAttr(buzz::QN_TO, message.to());
    result->AddAttr(buzz::QN_TYPE, buzz::STR_SET);
    buzz::XmlElement *session = new buzz::XmlElement(QN_GOOGLESESSION_SESSION, true);
    result->AddElement(session);
    switch (message.type()) {
    case cricket::SessionMessage::TYPE_INITIATE:
        session->AddAttr(buzz::QN_TYPE, "initiate");
        break;
    case cricket::SessionMessage::TYPE_ACCEPT:
        session->AddAttr(buzz::QN_TYPE, "accept");
        break;
    case cricket::SessionMessage::TYPE_MODIFY:
        session->AddAttr(buzz::QN_TYPE, "modify");
        break;
    case cricket::SessionMessage::TYPE_CANDIDATES:
        session->AddAttr(buzz::QN_TYPE, "candidates");
        break;
    case cricket::SessionMessage::TYPE_REJECT:
        session->AddAttr(buzz::QN_TYPE, "reject");
        break;
    case cricket::SessionMessage::TYPE_REDIRECT:
        session->AddAttr(buzz::QN_TYPE, "redirect");
        break;
    case cricket::SessionMessage::TYPE_TERMINATE:
        session->AddAttr(buzz::QN_TYPE, "terminate");
        break;
    }
    session->AddAttr(QN_ID, message.session_id().id_str());
    session->AddAttr(QN_INITIATOR, message.session_id().initiator());
    return result;
}

buzz::XmlElement *XmppCallSessionClient::TranslateCandidate(const cricket::Candidate &candidate) 
{
    buzz::XmlElement *result = new buzz::XmlElement(QN_GOOGLESESSION_CANDIDATE);
    result->AddAttr(QN_NAME, candidate.name());
    result->AddAttr(QN_ADDRESS, candidate.address().IPAsString());
    result->AddAttr(QN_PORT, candidate.address().PortAsString());
    result->AddAttr(QN_USERNAME, candidate.username());
    result->AddAttr(QN_PASSWORD, candidate.password());
    result->AddAttr(QN_PREFERENCE, candidate.preference_str());
    result->AddAttr(QN_PROTOCOL, candidate.protocol());
    result->AddAttr(buzz::QN_TYPE, candidate.type());
    result->AddAttr(QN_NETWORK, candidate.network_name());
    result->AddAttr(QN_GENERATION, candidate.generation_str());
    return result;
}

buzz::XmlElement *XmppCallSessionClient::TranslateInitiateAcceptModify(const cricket::XmppSessionMessage &message) 
{
    // Header info common to all message types
    buzz::XmlElement *result = TranslateHeader(message);
    buzz::XmlElement *session = result->FirstNamed(QN_GOOGLESESSION_SESSION);
    
    // Candidates
    assert(message.candidates().size() == 0);
    
    // Session Description
    buzz::XmlElement* description = TranslateSessionDescription(message.description());
    assert(description->Name().LocalPart() == "description");
    assert(description->Name().Namespace() == NS_PHONE);
    session->AddElement(description);
    
    if (message.redirect_cookie() != NULL) {
        const buzz::XmlElement* cookie =
            reinterpret_cast<XmlCookie*>(message.redirect_cookie())->elem();
        for (const buzz::XmlElement* elem = cookie->FirstElement(); elem; elem = elem->NextElement())
            session->AddElement(new buzz::XmlElement(*elem));
    }
    
    return result;
}

buzz::XmlElement *XmppCallSessionClient::TranslateCandidates(const cricket::XmppSessionMessage &message) 
{
    // Header info common to all message types
    buzz::XmlElement *result = TranslateHeader(message);
    buzz::XmlElement *session = result->FirstNamed(QN_GOOGLESESSION_SESSION);
    
    // Candidates
    std::vector<cricket::Candidate>::const_iterator it;
    for (it = message.candidates().begin(); it != message.candidates().end(); it++)
        session->AddElement(TranslateCandidate(*it));
    
    return result;
}

buzz::XmlElement *XmppCallSessionClient::TranslateRejectTerminate(const cricket::XmppSessionMessage &message) 
{
    // These messages are simple, and only have a header
    return TranslateHeader(message);
}

buzz::XmlElement *XmppCallSessionClient::TranslateRedirect(const cricket::XmppSessionMessage &message) 
{
    // Header info common to all message types
    buzz::XmlElement *result = TranslateHeader(message);
    buzz::XmlElement *session = result->FirstNamed(QN_GOOGLESESSION_SESSION);
    
    assert(message.candidates().size() == 0);
    assert(message.description() == NULL);
    
    assert(message.redirect_target().size() > 0);
    buzz::XmlElement* target = new buzz::XmlElement(QN_GOOGLESESSION_TARGET);
    target->AddAttr(QN_NAME, message.redirect_target());
    session->AddElement(target);
    
    buzz::XmlElement* cookie = new buzz::XmlElement(QN_GOOGLESESSION_COOKIE);
    session->AddElement(cookie);
    
    // If the message does not have a redirect cookie, then this is a redirect
    // initiated by us.  We will automatically add a regarding cookie.
    if (message.redirect_cookie() == NULL) {
        buzz::XmlElement* regarding = new buzz::XmlElement(QN_GOOGLESESSION_REGARDING);
        regarding->AddAttr(QN_NAME, GetClient()->jid().BareJid().Str());
        cookie->AddElement(regarding);
    } else {
        const buzz::XmlElement* cookie_elem =
            reinterpret_cast<const XmlCookie*>(message.redirect_cookie())->elem();
        const buzz::XmlElement* elem;
        for (elem = cookie_elem->FirstElement(); elem; elem = elem->NextElement())
            cookie->AddElement(new buzz::XmlElement(*elem));
    }
    
    return result;
}

buzz::XmlElement *XmppCallSessionClient::TranslateSessionDescription(const cricket::SessionDescription *_session_desc) 
{
    const XmppCallSessionDescription* session_desc =
        static_cast<const XmppCallSessionDescription*>(_session_desc);
    buzz::XmlElement* description = new buzz::XmlElement(QN_PHONE_DESCRIPTION, true);
    
    for (size_t i = 0; i < session_desc->codecs().size(); ++i) {
        buzz::XmlElement* payload_type = new buzz::XmlElement(QN_PHONE_PAYLOADTYPE, true);
        const XmppCallSessionDescription::Codec &codec = session_desc->codecs()[i];
        char buf[32];
        
        sprintf(buf, "%d", codec.id);
        payload_type->AddAttr(QN_PHONE_PAYLOADTYPE_ID, buf);
        payload_type->AddAttr(QN_PHONE_PAYLOADTYPE_NAME, codec.name.c_str());
        if (codec.rate != -1) {
            sprintf(buf, "%d", codec.rate);
            payload_type->AddAttr(QN_PHONE_PAYLOADTYPE_RATE, buf);
        }
        if (codec.channels != -1) {
            sprintf(buf, "%d", codec.channels);
            payload_type->AddAttr(QN_PHONE_PAYLOADTYPE_CHANNELS, buf);
        }
   
        description->AddElement(payload_type);
    }
    
    return description;
}

const cricket::SessionDescription *XmppCallSessionClient::CreateSessionDescription(const buzz::XmlElement *element) 
{
    XmppCallSessionDescription* desc = new XmppCallSessionDescription();
    
    const buzz::XmlElement* payload_type = element->FirstNamed(QN_PHONE_PAYLOADTYPE);
    int num_payload_types = 0;
    
    while (payload_type) {
        if (payload_type->HasAttr(QN_PHONE_PAYLOADTYPE_ID) &&
            payload_type->HasAttr(QN_PHONE_PAYLOADTYPE_NAME)) {
            int id = atoi(payload_type->Attr(QN_PHONE_PAYLOADTYPE_ID).c_str());
            int rate = -1;
            int channels = -1;
            int pref = 0;
            
            std::string name = payload_type->Attr(QN_PHONE_PAYLOADTYPE_NAME);
            if (payload_type->HasAttr(QN_PHONE_PAYLOADTYPE_RATE))
                int rate = atoi(payload_type->Attr(QN_PHONE_PAYLOADTYPE_RATE).c_str());
            if (payload_type->HasAttr(QN_PHONE_PAYLOADTYPE_CHANNELS))
                int channels = atoi(payload_type->Attr(QN_PHONE_PAYLOADTYPE_CHANNELS).c_str());
            desc->AddCodec(XmppCallSessionDescription::Codec(id, name, rate, channels, 0));
        }
        
        payload_type = payload_type->NextNamed(QN_PHONE_PAYLOADTYPE);
        num_payload_types += 1;
    }
    
    // For backward compatability, we can assume the other client is (an old
    // version of Talk) if it has no payload types at all.
    if (num_payload_types == 0) {
        desc->AddCodec(XmppCallSessionDescription::Codec(103, "ISAC", -1, -1, 1));
        desc->AddCodec(XmppCallSessionDescription::Codec(0, "PCMU", 8000, -1, 0));
    }
    
    return desc;
}

void XmppCallSessionClient::HandleIncomingMessage(const cricket::SessionMessage &m) 
{
    // In the case of an incoming initiate, there is no session yet, and one needs to be created.
    // The other cases have sessions already.
    
    XmppCall *call;
    switch (m.type()) {
    case cricket::SessionMessage::TYPE_INITIATE:
        call = CreateCall(m.name(), m.session_id(), true);
        break;
    
    case cricket::SessionMessage::TYPE_ACCEPT:
    case cricket::SessionMessage::TYPE_MODIFY:
    case cricket::SessionMessage::TYPE_CANDIDATES:
    case cricket::SessionMessage::TYPE_REJECT:
    case cricket::SessionMessage::TYPE_REDIRECT:
    case cricket::SessionMessage::TYPE_TERMINATE:
        call = GetCall(m.session_id());
        break;
    
    default:
        cricket::SessionResponseMessage *response =
            m.CreateErrorResponse(cricket::SessionResponseMessage::ERROR_UNKNOWN_METHOD);
        OnSendResponse(NULL, response);
        delete response;
        return;
    }
    
    if (call == NULL) {
        cricket::SessionResponseMessage *response =
            m.CreateErrorResponse(cricket::SessionResponseMessage::ERROR_UNKNOWN_SESSION);
        OnSendResponse(NULL, response);
        delete response;
    } else {
        call->OnIncomingMessage(m);
    }
}

void XmppCallSessionClient::OnSendResponse(XmppCall *,  // may be NULL
                                           const cricket::SessionResponseMessage *message) 
{
    const cricket::XmppSessionResponseMessage *m =
        static_cast<const cricket::XmppSessionResponseMessage *>(
                message->asClass(XMPP_CLIENT_CLASS_NAME));
    assert(m != NULL);
    
    const buzz::XmlElement *original_stanza = m->original_stanza();
    // We must have an original stanza in order to be able to set a
    // valid TO field.
    assert(original_stanza != NULL);
    
    buzz::XmlElement result(buzz::QN_IQ);
    result.AddAttr(buzz::QN_TYPE, buzz::STR_RESULT);
    assert(original_stanza->HasAttr(buzz::QN_FROM));
    result.AddAttr(buzz::QN_TO, original_stanza->Attr(buzz::QN_FROM));
    result.AddAttr(buzz::QN_ID, m->message_id());
    
    switch (m->type()) {
        case cricket::SessionResponseMessage::TYPE_OK:
            break;
        case cricket::SessionResponseMessage::TYPE_ERROR: {
            // Right now we don't change what we send but will eventually respond
            // with errors as necessary.
            break;
        }
        case cricket::SessionResponseMessage::TYPE_INITIATERESPONSE: {
            break;
        }
        default:
            assert(0);
    }
    
    XmppCallSessionSendTask *sender = new XmppCallSessionSendTask(GetClient(), this);
    sender->Send(&result);
    sender->Start();
}

void XmppCallSessionClient::OnOutgoingMessage(XmppCall *call, 
                                              const cricket::SessionMessage &message) 
{
    // Translate the message into an XMPP stanza
    const cricket::XmppSessionMessage *xmppMessage =
      static_cast<const cricket::XmppSessionMessage *>(
          message.asClass(XMPP_CLIENT_CLASS_NAME));
    assert(xmppMessage != NULL);
    
    buzz::XmlElement *result = NULL;
    switch (message.type()) {
    case cricket::SessionMessage::TYPE_INITIATE:
    case cricket::SessionMessage::TYPE_ACCEPT:
    case cricket::SessionMessage::TYPE_MODIFY:
        result = TranslateInitiateAcceptModify(*xmppMessage);
        break;
    
    case cricket::SessionMessage::TYPE_CANDIDATES:
        result = TranslateCandidates(*xmppMessage);
        break;
    
    case cricket::SessionMessage::TYPE_REJECT:
    case cricket::SessionMessage::TYPE_TERMINATE:
        result = TranslateRejectTerminate(*xmppMessage);
        break;
    
    case cricket::SessionMessage::TYPE_REDIRECT:
        result = TranslateRedirect(*xmppMessage);
        break;
    }
    
    if (result != NULL) {
        XmppCallSessionSendTask *sender = new XmppCallSessionSendTask(GetClient(), this);
        sender->Send(result);
        sender->Start();
    }
}

void XmppCallSessionClient::OnIncomingResponse(const buzz::XmlElement *original_stanza,
                                               const buzz::XmlElement *response_stanza) 
{
    const buzz::XmlElement *element = original_stanza->FirstNamed(QN_GOOGLESESSION_SESSION);
    if (element != NULL) {
        cricket::SessionID sessionId;
        if (element->HasAttr(QN_ID)) {
            sessionId.set_id_str(element->Attr(QN_ID));
        }
        if (element->HasAttr(QN_INITIATOR)) {
            sessionId.set_initiator(element->Attr(QN_INITIATOR));
        }
        XmppCall *call = GetCall(sessionId);
        if (call != NULL) {
            cricket::XmppSessionResponseMessage response;
            if (ParseIncomingResponse(response_stanza, response)) {
                call->OnIncomingResponse(response);
            }
        }
    }
}

void XmppCallSessionClient::OnFailedSend(const buzz::XmlElement *original_stanza,
                                         const buzz::XmlElement *failure_stanza) 
{
    cricket::XmppSessionMessage m;
    if (!ParseIncomingMessage(original_stanza, m))
        return;
    
    // Incoming signaling error. This means, as the result of trying
    // to send message m, and error was generated. In all cases, a
    // session should already exist
    
    XmppCall *call = NULL;
    switch (m.type()) {
    case cricket::SessionMessage::TYPE_INITIATE:
    case cricket::SessionMessage::TYPE_ACCEPT:
    case cricket::SessionMessage::TYPE_MODIFY:
    case cricket::SessionMessage::TYPE_CANDIDATES:
    case cricket::SessionMessage::TYPE_REJECT:
    case cricket::SessionMessage::TYPE_TERMINATE:
        call = GetCall(m.session_id());
        break;
    
    default:
        return;
    }
    
    if (call != NULL)
        call->OnIncomingError(m);
}
