#ifndef __XMPP_CALL_SESSION_SEND_TASK_H__
#define __XMPP_CALL_SESSION_SEND_TASK_H__

#include <talk/xmpp/constants.h>
#include <talk/xmpp/xmppengine.h>
#include <talk/xmpp/xmpptask.h>

#include "xmpp_call_session_client.h"

// The job of this task is to send an IQ stanza out (after stamping it with
// an ID attribute) and then wait for a response.  If not response happens
// within 5 seconds, it will signal failure on a SessionClient.  If an error
// happens it will also signal failure.  If, however, the send succeeds this
// task will quietly go away.

// It is safe for this to hold on to the session client.  In the case where
// the xmpp client goes away, this task will automatically be aborted.  The
// session_client is guaranteed to outlive the xmpp session.
class XmppCallSessionSendTask : public buzz::XmppTask {
public:
    XmppCallSessionSendTask(Task *parent, XmppCallSessionClient *session_client)
        : buzz::XmppTask(parent, buzz::XmppEngine::HL_SINGLE),
          session_client_(session_client),
          timed_out_(false) 
    { 
    }
    
    void Send(const buzz::XmlElement* stanza) 
    {
        assert(stanza_.get() == NULL);
        stanza_.reset(new buzz::XmlElement(*stanza));
        if (stanza_->HasAttr(buzz::QN_ID)) {
            set_task_id(stanza_->Attr(buzz::QN_ID));
        } else {
            stanza_->SetAttr(buzz::QN_ID, task_id());
        }
    }

protected:
    /* This gets called by the task runner every 500 msec */
    void Poll() {
        if (ElapsedTime() > (15 * 1000 * 10000)) {  // 15 secs
            timed_out_ = true;
            Wake();
        }
    }

    int ProcessStart() 
    {
        SendStanza(stanza_.get());
        return STATE_RESPONSE;
    }

    int ProcessResponse() {
        if (timed_out_) {
            session_client_->OnFailedSend(stanza_.get(), NULL);
            return STATE_DONE;
        }

        const buzz::XmlElement* next = NextStanza();
        if (next == NULL)
            return STATE_BLOCKED;
        
        if (next->Attr(buzz::QN_TYPE) == buzz::STR_RESULT) {
            session_client_->OnIncomingResponse(stanza_.get(), next);
            return STATE_DONE;
        } 
        else {
            session_client_->OnFailedSend(stanza_.get(), next);
            return STATE_DONE;
        }
    }

    virtual bool HandleStanza(const buzz::XmlElement *stanza) {
        if (!MatchResponseIq(stanza,
                    buzz::Jid(stanza_->Attr(buzz::QN_TO)), task_id()))
            return false;
        if (stanza->Attr(buzz::QN_TYPE) == buzz::STR_RESULT ||
            stanza->Attr(buzz::QN_TYPE) == buzz::STR_ERROR) {
            QueueStanza(stanza);
            return true;
        }
        return false;
    }

private:
    XmppCallSessionClient *session_client_;
    buzz::scoped_ptr<buzz::XmlElement> stanza_;
    bool timed_out_;
};

#endif
