/*
 * Copyright (c) 2005 INdT.
 * @author Andre Moreira Magalhaes <andre.magalhaes@indt.org.br> 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "xmpp_chat_session_client.h"

#include "xmpp_chat.h"

#include <talk/base/logging.h>
#include <talk/xmpp/constants.h>
#include <talk/xmpp/xmppclient.h>

namespace {
    
const std::string STR_CHAT("chat");

}

using namespace buzz;

XmppChatSessionClient::XmppChatSessionClient(buzz::Task *parent) 
    : buzz::XmppTask(parent, XmppEngine::HL_TYPE) 
{
}

XmppChatSessionClient::~XmppChatSessionClient()
{
}

bool XmppChatSessionClient::HandleStanza(const buzz::XmlElement *stanza) 
{
    if (stanza->Name() != buzz::QN_MESSAGE)
        return false;
    if (stanza->HasAttr(buzz::QN_TYPE)) {
        std::string type = stanza->Attr(buzz::QN_TYPE);
        if (type != STR_CHAT && type != STR_ERROR)
            return false;
    }
    QueueStanza(stanza);
    return true;
}

XmppChat *XmppChatSessionClient::CreateChat(const buzz::Jid &remote_jid, 
                                            bool incoming)
{
    cricket::CritScope s(&crit_);

    XmppChat *chat = new XmppChat(this, remote_jid);
    chats_[remote_jid.Str()] = chat;
    SignalChatCreate(chat, incoming);
    return chat;
}
    
void XmppChatSessionClient::DestroyChat(XmppChat *chat)
{
    cricket::CritScope s(&crit_);
    
    SignalChatDestroy(chat);
    
    std::map<std::string, XmppChat *>::iterator it = chats_.find(chat->remote_jid().Str());
    if (it != chats_.end())
        chats_.erase(it);
    delete chat;
}

XmppMessage *XmppChatSessionClient::CreateMessage(const buzz::XmlElement *element)
{
    XmppMessage *message = new XmppMessage;
    message->from = element->Attr(buzz::QN_FROM);
    message->to = element->Attr(buzz::QN_TO);
    message->type = element->Attr(buzz::QN_TYPE);

    const buzz::XmlElement *subject = element->FirstNamed(buzz::QN_SUBJECT);
    if (subject)
        message->subject = subject->BodyText();
    const buzz::XmlElement *body = element->FirstNamed(buzz::QN_BODY);
    if (body)
        message->body = body->BodyText();
    const buzz::XmlElement *thread = element->FirstNamed(buzz::QN_THREAD);
    if (thread)
        message->thread = thread->BodyText();
    return message;
}

buzz::XmlElement *XmppChatSessionClient::TranslateMessage(const XmppMessage *message)
{
    XmlElement *result = new XmlElement(buzz::QN_MESSAGE);
    result->AddAttr(buzz::QN_ID, message->id);
    result->AddAttr(buzz::QN_FROM, message->from);
    result->AddAttr(buzz::QN_TO, message->to);
    result->AddAttr(buzz::QN_TYPE, message->type);
    if (!message->subject.empty()) {
        result->AddElement(new XmlElement(buzz::QN_SUBJECT));
        result->AddText(message->subject, 1);
    }
    if (!message->body.empty()) {
        result->AddElement(new XmlElement(buzz::QN_BODY));
        result->AddText(message->body, 1);
    }
    if (!message->thread.empty()) {
        result->AddElement(new XmlElement(buzz::QN_THREAD));
        result->AddText(message->thread, 1);
    }
    return result;
}

int XmppChatSessionClient::ProcessStart() 
{
    const XmlElement *stanza = NextStanza();
    if (stanza == NULL)
        return STATE_BLOCKED;

    crit_.Enter();
    
    XmppChat *chat;
    buzz::Jid remote_jid = buzz::Jid(stanza->Attr(buzz::QN_FROM));
    std::map<std::string, XmppChat *>::iterator it = chats_.find(remote_jid.Str());
    if (it != chats_.end()) 
        chat = (*it).second;
    else {
        std::map<std::string, XmppChat *>::iterator it = chats_.find(remote_jid.BareJid().Str());
        if (it != chats_.end()) {
            chat = (*it).second;
            /* chat is not established yet
             * establish it replacing chat bare jid with jid.
             */
            chat->SetRemoteJid(remote_jid);
            chats_.erase(it);
            chats_[remote_jid.Str()] = chat;
        }
        else
            chat = CreateChat(remote_jid, true);
    }
    
    crit_.Leave();
 
    XmppMessage *message = CreateMessage(stanza);
    if (message->type == "error") {
        const buzz::XmlElement *error = stanza->FirstNamed(buzz::QN_ERROR);
        if (error) {
            int error_code = atoi(error->Attr(buzz::QN_CODE).c_str());
            chat->OnError(*message, error_code);
        }
        else {
            // this shouldn't happen. ignore for now
        }
    }
    else
        chat->OnMessage(*message);
    delete message;

    return STATE_START;
}
