/*
 * Tapioca XMPP Plugin
 * Copyright (c) 2005 INdT.
 * @author Andre Moreira Magalhaes <andre.magalhaes@indt.org.br> 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "xmpp_client.h"

#include "xmpp_call_session_client.h"
#include "xmpp_call.h"
#include "xmpp_call_c.h"
#include "xmpp_chat_session_client.h"
#include "xmpp_chat.h"
#include "xmpp_chat_c.h"
#include "xmpp_roster.h"
#include "xmpp_presence_push.h"

#include "xmppsocket.h"
#include "presenceouttask.h"
#include "status.h"

#include <talk/base/logging.h>
#include <talk/base/network.h>
#include <talk/base/socketaddress.h>
#include <talk/base/thread.h>
#include <talk/p2p/base/helpers.h>
#include <talk/p2p/base/sessionmanager.h>
#include <talk/p2p/client/basicportallocator.h>
#include <talk/xmpp/constants.h>
#include <talk/xmpp/xmppclientsettings.h>

#include <iomanip>

class DebugLog : public sigslot::has_slots<> {
public:
  DebugLog() :
    debug_input_buf_(NULL), debug_input_len_(0), debug_input_alloc_(0),
    debug_output_buf_(NULL), debug_output_len_(0), debug_output_alloc_(0),
    censor_password_(false)
      {}
  char * debug_input_buf_;
  int debug_input_len_;
  int debug_input_alloc_;
  char * debug_output_buf_;
  int debug_output_len_;
  int debug_output_alloc_;
  bool censor_password_;

  void Input(const char * data, int len) {
    if (debug_input_len_ + len > debug_input_alloc_) {
      char * old_buf = debug_input_buf_;
      debug_input_alloc_ = 4096;
      while (debug_input_alloc_ < debug_input_len_ + len) {
        debug_input_alloc_ *= 2;
      }
      debug_input_buf_ = new char[debug_input_alloc_];
      memcpy(debug_input_buf_, old_buf, debug_input_len_);
      delete[] old_buf;
    }
    memcpy(debug_input_buf_ + debug_input_len_, data, len);
    debug_input_len_ += len;
    DebugPrint(debug_input_buf_, &debug_input_len_, false);
  }

  void Output(const char * data, int len) {
    if (debug_output_len_ + len > debug_output_alloc_) {
      char * old_buf = debug_output_buf_;
      debug_output_alloc_ = 4096;
      while (debug_output_alloc_ < debug_output_len_ + len) {
        debug_output_alloc_ *= 2;
      }
      debug_output_buf_ = new char[debug_output_alloc_];
      memcpy(debug_output_buf_, old_buf, debug_output_len_);
      delete[] old_buf;
    }
    memcpy(debug_output_buf_ + debug_output_len_, data, len);
    debug_output_len_ += len;
    DebugPrint(debug_output_buf_, &debug_output_len_, true);
  }

  static bool
  IsAuthTag(const char * str, size_t len) {
    if (str[0] == '<' && str[1] == 'a' &&
                         str[2] == 'u' &&
                         str[3] == 't' &&
                         str[4] == 'h' &&
                         str[5] <= ' ') {
      std::string tag(str, len);

      if (tag.find("mechanism") != std::string::npos)
        return true;

    }
    return false;
  }

  void
  DebugPrint(char * buf, int * plen, bool output) {
    int len = *plen;
    if (len > 0) {
      time_t tim = time(NULL);
      struct tm * now = localtime(&tim);
      char *time_string = asctime(now);
      if (time_string) {
        size_t time_len = strlen(time_string);
        if (time_len > 0) {
          time_string[time_len-1] = 0;    // trim off terminating \n
        }
      }
      LOG(INFO) << (output ? "SEND >>>>>>>>>>>>>>>>>>>>>>>>>" : "RECV <<<<<<<<<<<<<<<<<<<<<<<<<")
        << " : " << time_string;

      bool indent;
      int start = 0, nest = 3;
      for (int i = 0; i < len; i += 1) {
        if (buf[i] == '>') {
          if ((i > 0) && (buf[i-1] == '/')) {
            indent = false;
          } else if ((start + 1 < len) && (buf[start + 1] == '/')) {
            indent = false;
            nest -= 2;
          } else {
            indent = true;
          }

          // Output a tag
          LOG(INFO) << std::setw(nest) << " " << std::string(buf + start, i + 1 - start);

          if (indent)
            nest += 2;

          // Note if it's a PLAIN auth tag
          if (IsAuthTag(buf + start, i + 1 - start)) {
            censor_password_ = true;
          }

          // incr
          start = i + 1;
        }

        if (buf[i] == '<' && start < i) {
          if (censor_password_) {
            LOG(INFO) << std::setw(nest) << " " << "## TEXT REMOVED ##";
            censor_password_ = false;
          }
          else {
            LOG(INFO) << std::setw(nest) << " " << std::string(buf + start, i - start);
          }
          start = i;
        }
      }
      len = len - start;
      memcpy(buf, buf + start, len);
      *plen = len;
    }
  }

};

static DebugLog debug_log_;

XmppClient::XmppClient() 
    : pump_(new XmppPump(this)), 
      media_engine_(NULL),
      worker_thread_(NULL),
      network_manager_(NULL),
      port_allocator_(NULL),
      session_manager_(NULL),
      call_client_(NULL),
      chat_client_(NULL),
      receiver_(NULL),
      presence_out_(NULL),
      presence_push_(NULL),
      roster_(NULL),
      state_change_cb_(NULL),
      state_change_cb_data_(NULL),
      incoming_call_cb_(NULL),
      incoming_call_cb_data_(NULL),
      incoming_chat_cb_(NULL),
      incoming_chat_cb_data_(NULL)
{ 
    pump_->client()->SignalLogInput.connect(&debug_log_, &DebugLog::Input);
    pump_->client()->SignalLogOutput.connect(&debug_log_, &DebugLog::Output);
    LogMessage::LogThreads();
    LogMessage::LogTimestamps();
}

XmppClient::~XmppClient()
{
}

void XmppClient::Connect(const buzz::XmppClientSettings &xcs)
{
    pump_->DoLogin(xcs, new XmppSocket(xcs.use_tls()));
}

void XmppClient::Disconnect()
{
    pump_->DoDisconnect();
}

void XmppClient::SendRaw(const std::string &data)
{
    pump_->client()->SendRaw(data);
}

void XmppClient::OnStateChange(buzz::XmppEngine::State state)
{
    if (state == buzz::XmppEngine::STATE_OPEN) {
        std::string client_unique = pump_->client()->jid().Str();
        cricket::InitRandom(client_unique.c_str(), client_unique.size());
    
        InitChatClient();
        InitCallClient();
        InitRoster();
        InitPresence();
    }

    if (state_change_cb_)
        state_change_cb_((xmpp_client_t) this, state_change_cb_data_, state, pump_->client()->GetError());
}

void XmppClient::InitRoster()
{
    roster_ = new XmppRoster(pump_->client());
    roster_->Start();
}

void XmppClient::InitPresence()
{
    presence_push_ = new XmppPresencePush(pump_->client());
    presence_out_ = new buzz::PresenceOutTask(pump_->client());
    presence_out_->Start();
}

void XmppClient::InitChatClient()
{
    chat_client_ = new XmppChatSessionClient(pump_->client());
    chat_client_->SignalChatCreate.connect(this, &XmppClient::OnChatCreate);
    chat_client_->Start();
}

void XmppClient::InitCallClient()
{
    call_client_ = new XmppCallSessionClient(pump_->client());
    call_client_->SignalCallCreate.connect(this, &XmppClient::OnCallCreate);
    call_client_->Start();
}

void XmppClient::OnChatCreate(XmppChat *chat,
                              bool incoming)
{
    if (incoming && incoming_chat_cb_) {
        incoming_chat_cb_((xmpp_client_t) this, incoming_chat_cb_data_, 
                    chat->remote_jid().Str().c_str(), (xmpp_chat_t) chat);
    }
}

void XmppClient::OnCallCreate(XmppCall *call,
                              bool incoming)
{
    if (incoming && incoming_call_cb_) {
        incoming_call_cb_((xmpp_client_t) this, incoming_call_cb_data_, 
                    call->id().initiator().c_str(), (xmpp_call_t) call);
    }
}
