/*
 * Tapioca XMPP Plugin
 * Copyright (c) 2005 INdT.
 * @author Luiz Augusto von Dentz <luiz.dentz@indt.org.br> 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif 

#include "xmpp_contact_list.h"

#include <tapioca/core/tapioca_enums.h>
#include <tapioca/core/tapioca_errors.h>
#include <tapioca/core/tapioca_utils.h>

#include <string.h>

struct _XmppContactListPrivate {
    gboolean disposed;
    TapiocaContact **list;

    xmpp_roster_t roster;
};

static void     xmpp_contact_list_dispose                (GObject *object);
    
static void     xmpp_contact_list_request_contacts       (TapiocaContactList *tcl);
static void     xmpp_contact_list_request_contact_info   (TapiocaContactList *tcl,
                                                          const gchar *address);
static void     xmpp_contact_list_add_contact            (TapiocaContactList *tcl,
                                                          const gchar *address,
                                                          const gchar *alias,
                                                          const gchar *group);
static void     xmpp_contact_list_remove_contact         (TapiocaContactList *tcl,
                                                          const gchar *address);
static void     xmpp_contact_list_authorize_contact      (TapiocaContactList *tcl,
                                                          const gchar *address,
                                                          const gchar *group);
static void     xmpp_contact_list_deny_contact           (TapiocaContactList *tcl,
                                                          const gchar *address);
static void     authorization_requested_cb               (xmpp_roster_t roster,
                                                          gpointer data,
                                                          const gchar *jid);
static void     contact_denied_cb                        (xmpp_roster_t roster,
                                                          gpointer data,
                                                          const gchar *jid);
static void     contact_authorized_cb                    (xmpp_roster_t roster,
                                                          gpointer data,
                                                          const gchar *jid);
static void     contact_retrieved_cb                     (xmpp_roster_t roster,
                                                          gpointer data,
                                                          const gchar *jid,
                                                          const gchar *name,
                                                          const gchar *group,
                                                          const gchar *subscription);
G_DEFINE_TYPE(XmppContactList, xmpp_contact_list, TAPIOCA_TYPE_CONTACT_LIST)

static void 
xmpp_contact_list_class_init (XmppContactListClass *klass)
{
    GObjectClass *gobject_class;
    TapiocaContactListClass *tapioca_contact_list_class;

    gobject_class = (GObjectClass *) klass;
    tapioca_contact_list_class = (TapiocaContactListClass*) klass;
    xmpp_contact_list_parent_class = g_type_class_peek_parent (klass);
    
    g_type_class_add_private (klass, sizeof (XmppContactListPrivate));

    gobject_class->dispose = xmpp_contact_list_dispose;

    tapioca_contact_list_class->request_contacts = xmpp_contact_list_request_contacts;
    tapioca_contact_list_class->request_contact_info = xmpp_contact_list_request_contact_info;
    tapioca_contact_list_class->add_contact = xmpp_contact_list_add_contact;
    tapioca_contact_list_class->remove_contact = xmpp_contact_list_remove_contact;
    tapioca_contact_list_class->authorize_contact = xmpp_contact_list_authorize_contact;
    tapioca_contact_list_class->deny_contact = xmpp_contact_list_deny_contact;
}

static void 
xmpp_contact_list_init (XmppContactList *xc)
{
    xc->priv = XMPP_CONTACT_LIST_GET_PRIVATE (xc);
    xc->priv->list = NULL;
}

static void
xmpp_contact_list_dispose (GObject *object)
{
    XmppContactList *xc = XMPP_CONTACT_LIST (object);

    if (xc->priv->disposed) {
        /* If dispose did already run, return. */
        return;
    }
    /* Make sure dispose does not run twice. */
    xc->priv->disposed = TRUE;
   
    G_OBJECT_CLASS (xmpp_contact_list_parent_class)->dispose (object);
}

XmppContactList *
xmpp_contact_list_new (void)
{
    return XMPP_CONTACT_LIST (g_object_new (XMPP_TYPE_CONTACT_LIST, NULL));
}

void
xmpp_contact_list_initialize (XmppContactList *xcl, xmpp_roster_t *roster)
{
    g_debug ("initializing contact list");

    g_return_if_fail (roster != NULL);
    
    xcl->priv->roster = roster;

    xmpp_roster_set_authorization_requested_cb (roster, authorization_requested_cb, xcl);
    xmpp_roster_set_contact_authorized_cb (roster, contact_authorized_cb, xcl);
    xmpp_roster_set_contact_denied_cb (roster, contact_denied_cb, xcl);
    xmpp_roster_set_contact_retrieved_cb (roster, contact_retrieved_cb, xcl);
}

static void
xmpp_contact_list_request_contacts (TapiocaContactList *tcl)
{
    XmppContactList *xcl = XMPP_CONTACT_LIST(tcl);
    
    g_debug ("xmpp_contact_list_request_contacts called!");

    g_return_if_fail (xcl->priv->roster != NULL);

    xmpp_roster_request_contacts (xcl->priv->roster);
}

static void
xmpp_contact_list_request_contact_info (TapiocaContactList *tcl,
                                        const gchar *address)
{
    XmppContactList *xcl = XMPP_CONTACT_LIST(tcl);
    
    g_debug ("xmpp_contact_list_request_contact_info called! address=%s", address);

    g_return_if_fail (xcl->priv->roster != NULL);

    xmpp_roster_request_info (xcl->priv->roster, address);
}

static void
xmpp_contact_list_add_contact (TapiocaContactList *tcl,
                               const gchar *address,
                               const gchar *alias,
                               const gchar *group)
{
    XmppContactList *xcl = XMPP_CONTACT_LIST(tcl);
    
    g_debug ("xmpp_contact_list_add_contact called! address=%s, alias=%s, group=%s",
        address, alias, group);

    g_return_if_fail (xcl->priv->roster != NULL);

    xmpp_roster_add_contact (xcl->priv->roster, address, alias, group);
}

static void
xmpp_contact_list_remove_contact (TapiocaContactList *tcl,
                                  const gchar *address)
{
    XmppContactList *xcl = XMPP_CONTACT_LIST(tcl);
    
    g_debug ("xmpp_contact_list_remove_contact called! address=%s", address);

    g_return_if_fail (xcl->priv->roster != NULL);

    xmpp_roster_remove_contact (xcl->priv->roster, address);
}

static void
xmpp_contact_list_authorize_contact (TapiocaContactList *tcl,
                                     const gchar *address,
                                     const gchar *group)
{
    XmppContactList *xcl = XMPP_CONTACT_LIST(tcl);
    
    g_debug ("xmpp_contact_list_authorize_contact called!");

    g_return_if_fail (xcl->priv->roster != NULL);
    
    xmpp_roster_authorize_contact (xcl->priv->roster, address);
}

static void
xmpp_contact_list_deny_contact (TapiocaContactList *tcl,
                                const gchar *address)
{
    XmppContactList *xcl = XMPP_CONTACT_LIST(tcl);
    
    g_debug ("xmpp_contact_list_deny_contact called!");

    g_return_if_fail (xcl->priv->roster != NULL);

    xmpp_roster_deny_contact (xcl->priv->roster, address);
}

static void
authorization_requested_cb (xmpp_roster_t roster,
                            gpointer data,
                            const gchar *jid)
{
    TapiocaContactList *tcl = TAPIOCA_CONTACT_LIST (data);

    g_debug ("requesting authorization by %s", jid);

    tapioca_contact_list_signal_authorization_requested (tcl, jid);
}

static void
contact_denied_cb (xmpp_roster_t roster,
                   gpointer data,
                   const gchar *jid)
{
    TapiocaContactList *tcl = TAPIOCA_CONTACT_LIST (data);

    g_debug ("denying contact %s", jid);

    tapioca_contact_list_signal_contact_denied (tcl, jid);
}

static void
contact_authorized_cb (xmpp_roster_t roster,
                       gpointer data,
                       const gchar *jid)
{
    TapiocaContactList *tcl = TAPIOCA_CONTACT_LIST (data);

    g_debug ("authorizing contact %s", jid);

    tapioca_contact_list_signal_contact_authorized (tcl, jid);
}

static void
contact_retrieved_cb (xmpp_roster_t roster,
                      gpointer data,
                      const gchar *jid,
                      const gchar *name,
                      const gchar *group,
                      const gchar *subscription)
{
    TapiocaContactList *tcl = TAPIOCA_CONTACT_LIST (data);
    TapiocaContactListSubscriptionState subscription_type = 
        TAPIOCA_CONTACT_LIST_SUBSCRIPTION_STATE_NONE;
    
    if (!strcmp(subscription, "none"))
        subscription_type = TAPIOCA_CONTACT_LIST_SUBSCRIPTION_STATE_NONE;
    else if (!strcmp(subscription, "both"))
        subscription_type = TAPIOCA_CONTACT_LIST_SUBSCRIPTION_STATE_BOTH;
    else if (!strcmp(subscription, "from"))
        subscription_type = TAPIOCA_CONTACT_LIST_SUBSCRIPTION_STATE_FROM;
    else if (!strcmp(subscription, "to"))
        subscription_type = TAPIOCA_CONTACT_LIST_SUBSCRIPTION_STATE_TO;

    g_debug ("retrieving contact: %s, %s, %s, %s", jid, name, group, subscription);

    tapioca_contact_list_signal_contacts_retrieved (tcl, jid, name, group,
        subscription_type);
}
