/*
 * Tapioca XMPP Plugin
 * Copyright (c) 2005 INdT.
 * @author Andre Moreira Magalhaes <andre.magalhaes@indt.org.br> 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif 

#include "xmpp_presence.h"

#include <tapioca/core/tapioca_enums.h>
#include <tapioca/core/tapioca_errors.h>
#include <tapioca/core/tapioca_utils.h>

#include <string.h>

struct _XmppPresencePrivate {
    xmpp_presence_out_t presence_out;
    xmpp_presence_push_t presence_push;
    gboolean disposed;
};

static void     xmpp_presence_dispose                (GObject *object);

static gboolean xmpp_presence_subscribe              (TapiocaPresence *tp,
                                                      guint status,
                                                      const gchar *status_desc,
                                                      GError **error);
static gboolean xmpp_presence_unsubscribe            (TapiocaPresence *tp,
                                                      GError **error);
static void     xmpp_presence_get_supported_statuses (TapiocaPresence *tp,
                                                      GArray **statuses);
static gboolean xmpp_presence_set_status             (TapiocaPresence *tp,
                                                      guint status,
                                                      const gchar *status_desc,
                                                      GError **error);
static void     xmpp_presence_request_status         (TapiocaPresence *tp,
                                                      gchar **list);
static void     xmpp_presence_release_status         (TapiocaPresence *tp,
                                                      gchar **list);
static void     status_updated_cb                    (xmpp_presence_push_t presence_push,
                                                      gpointer data, 
                                                      const gchar *jid,
                                                      guint status,
                                                      const gchar *status_desc,
                                                      gboolean phone_capability);
static void     emit_status_updated                  (TapiocaPresence *tp,
                                                      const gchar *jid,
                                                      guint status,
                                                      const gchar *status_desc,
                                                      gboolean phone_capability);
static guint    get_status_from_tapioca_status       (guint status);
    
G_DEFINE_TYPE(XmppPresence, xmpp_presence, TAPIOCA_TYPE_PRESENCE)

static void 
xmpp_presence_class_init (XmppPresenceClass *klass)
{
    GObjectClass *gobject_class;
    TapiocaPresenceClass *tapioca_presence_class;

    gobject_class = (GObjectClass *) klass;
    tapioca_presence_class = (TapiocaPresenceClass*) klass;
    xmpp_presence_parent_class = g_type_class_peek_parent (klass);
    
    g_type_class_add_private (klass, sizeof (XmppPresencePrivate));

    gobject_class->dispose = xmpp_presence_dispose;
    
    tapioca_presence_class->subscribe = xmpp_presence_subscribe;
    tapioca_presence_class->unsubscribe = xmpp_presence_unsubscribe;
    tapioca_presence_class->get_supported_statuses = xmpp_presence_get_supported_statuses;
    tapioca_presence_class->set_status = xmpp_presence_set_status;
    tapioca_presence_class->request_status = xmpp_presence_request_status;
    tapioca_presence_class->release_status = xmpp_presence_release_status;
}

static void 
xmpp_presence_init (XmppPresence *xc)
{
    xc->priv = XMPP_PRESENCE_GET_PRIVATE (xc);
    xc->priv->presence_out = NULL;
    xc->priv->presence_push = NULL;
    xc->priv->disposed = FALSE;
}

static void
xmpp_presence_dispose (GObject *object)
{
    XmppPresence *xc = XMPP_PRESENCE (object);

    if (xc->priv->disposed) {
        /* If dispose did already run, return. */
        return;
    }

    /* Make sure dispose does not run twice. */
    xc->priv->disposed = TRUE;
   
    G_OBJECT_CLASS (xmpp_presence_parent_class)->dispose (object);
}

XmppPresence *
xmpp_presence_new (void)
{
    return XMPP_PRESENCE (g_object_new (XMPP_TYPE_PRESENCE, NULL));
}

void
xmpp_presence_initialize (XmppPresence *xp,
                          xmpp_presence_out_t presence_out,
                          xmpp_presence_push_t presence_push)
{
    g_debug ("initializing presence");

    xp->priv->presence_out = presence_out;
    xp->priv->presence_push = presence_push;
    xmpp_presence_push_set_status_updated_cb (presence_push,
                                              status_updated_cb,
                                              xp);
}

static gboolean 
xmpp_presence_subscribe (TapiocaPresence *tp,
                         guint status,
                         const gchar *status_desc,
                         GError **error)
{
    XmppPresence *xp = XMPP_PRESENCE (tp);

    if (!xp->priv->presence_out) {
        g_set_error (error, TAPIOCA_ERROR, TAPIOCA_ERROR_GENERIC, "Presence not available yet");
        return FALSE;
    }
            
    g_debug ("subscribing with status %d", status);
    xmpp_presence_out_set_status (xp->priv->presence_out,
            get_status_from_tapioca_status (status),
            status_desc);
    tapioca_presence_signal_subscribed (tp);

    return TRUE;
}

static gboolean 
xmpp_presence_unsubscribe (TapiocaPresence *tp,
                           GError **error)
{
    XmppPresence *xp = XMPP_PRESENCE (tp);

    if (!xp->priv->presence_out) {
        g_set_error (error, TAPIOCA_ERROR, TAPIOCA_ERROR_GENERIC, "Presence not available yet");
        return FALSE;
    }
 
    g_debug ("unsubscribing");
    xmpp_presence_out_set_status (xp->priv->presence_out, 1, NULL);
    tapioca_presence_signal_unsubscribed (tp);

    return TRUE;
}

static void 
xmpp_presence_get_supported_statuses (TapiocaPresence *tp,
                                      GArray **statuses)
{
    guint i;
    guint supported_statuses[] = { 
        TAPIOCA_PRESENCE_STATUS_OFFLINE,
        TAPIOCA_PRESENCE_STATUS_AVAILABLE,
        TAPIOCA_PRESENCE_STATUS_AWAY,
        TAPIOCA_PRESENCE_STATUS_XA,
        TAPIOCA_PRESENCE_STATUS_DND 
    };
    guint count = sizeof (supported_statuses) / sizeof (guint);

    *statuses = g_array_sized_new (FALSE, FALSE, sizeof (guint), 5);
    for (i = 0; i < count; ++i) {
        *statuses = g_array_append_val (*statuses, supported_statuses[i]);
    }
}

static gboolean 
xmpp_presence_set_status (TapiocaPresence *tp,
                          guint status,
                          const char *status_desc,
                          GError **error)
{
    XmppPresence *xp = XMPP_PRESENCE (tp);
    
    if (!xp->priv->presence_out) {
        g_set_error (error, TAPIOCA_ERROR, TAPIOCA_ERROR_GENERIC, "Presence not available yet");
        return FALSE;
    }

    if (status_desc)
        g_debug ("setting status to %d with description = %s ...", status, status_desc);
    else
        g_debug ("setting status to %d ...", status);

    xmpp_presence_out_set_status (xp->priv->presence_out, 
            get_status_from_tapioca_status (status),
            status_desc);

    return TRUE;
}

static void 
xmpp_presence_request_status (TapiocaPresence *tp,
                              gchar **list)
{
    
    XmppPresence *xp = XMPP_PRESENCE (tp);
    guint i = 0;

    if (!xp->priv->presence_push) {
        return;
    }
    
    for (; list[i]; ++i) {
        xmpp_presence_push_request_status (xp->priv->presence_push, list[i]);
    }
}

static void 
xmpp_presence_release_status (TapiocaPresence *tp,
                              gchar **list)
{
    // do nothing
}

static void 
status_updated_cb (xmpp_presence_push_t presence_push,
                   gpointer data, 
                   const gchar *jid,
                   guint status,
                   const gchar *status_desc,
                   gboolean phone_capability)
{
    TapiocaPresence *tp = TAPIOCA_PRESENCE (data);
   
    /* check if jid "node@domain/resource" our
     * bare jid "node@domain" is on watch list
     */
    if (tapioca_presence_in_watch_list (tp, jid)) {
        emit_status_updated (tp, jid, status, status_desc, phone_capability);
    }
    else {
        gchar *offset = strchr (jid, '/');
        gchar *bare_jid;
        if (!offset)
            return;
        
        bare_jid = g_strndup (jid, offset - jid);
        if (tapioca_presence_in_watch_list (tp, bare_jid)) {
            emit_status_updated (tp, jid, status, status_desc, phone_capability);
        }
        g_free (bare_jid);
    }
}

static void 
emit_status_updated (TapiocaPresence *tp,
                     const gchar *jid,
                     guint status,
                     const gchar *status_desc,
                     gboolean phone_capability)
{
    switch (status) {
        case 0:
        case 1:
            tapioca_presence_signal_status_updated (tp, jid, TAPIOCA_PRESENCE_STATUS_OFFLINE, status_desc, phone_capability);
            break;
        case 2:
            tapioca_presence_signal_status_updated (tp, jid, TAPIOCA_PRESENCE_STATUS_XA, status_desc, phone_capability);
            break;
        case 3:
            tapioca_presence_signal_status_updated (tp, jid, TAPIOCA_PRESENCE_STATUS_AWAY, status_desc, phone_capability);
            break;
        case 4:
            tapioca_presence_signal_status_updated (tp, jid, TAPIOCA_PRESENCE_STATUS_DND, status_desc, phone_capability);
            break;
        case 5:
        case 6:
            tapioca_presence_signal_status_updated (tp, jid, TAPIOCA_PRESENCE_STATUS_AVAILABLE, status_desc, phone_capability);
            break;
        default:
            g_debug ("received unknown status %d for contact %s", status, jid);
            break; 
    }
}

static guint 
get_status_from_tapioca_status (guint status)
{
    switch (status) {
        case TAPIOCA_PRESENCE_STATUS_OFFLINE:
            return 1;
        case TAPIOCA_PRESENCE_STATUS_AVAILABLE:
            return 5;
        case TAPIOCA_PRESENCE_STATUS_AWAY:
            return 3;
        case TAPIOCA_PRESENCE_STATUS_XA:
            return 2;
        case TAPIOCA_PRESENCE_STATUS_DND:
            return 4;
    }
    return 0;
}
