/*
 * Copyright (c) 2005 INdT.
 * @author Luiz Augusto von Dentz <luiz.dentz@indt.org.br> 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "xmpp_roster.h"

#include <talk/p2p/base/helpers.h>
#include <talk/base/logging.h>
#include <talk/xmpp/constants.h>
#include <talk/xmpp/xmppclient.h>


XmppRoster::XmppRoster(buzz::Task *parent) 
    : buzz::XmppTask(parent, buzz::XmppEngine::HL_TYPE) 
{
}

XmppRoster::~XmppRoster()
{
}

void XmppRoster::AddContact(const buzz::Jid &jid, const std::string &name, 
    const std::string &group)
{
    buzz::XmlElement *result = new buzz::XmlElement(buzz::QN_IQ);
    buzz::XmlElement *query = new buzz::XmlElement(buzz::QN_ROSTER_QUERY);
    buzz::XmlElement *item = new buzz::XmlElement(buzz::QN_ROSTER_ITEM);
    buzz::XmlElement *item_group = new buzz::XmlElement(buzz::QN_ROSTER_GROUP);
    buzz::XmlElement *subscribe = new buzz::XmlElement(buzz::QN_PRESENCE);
    std::string id_str;
    std::stringstream st;
    st << cricket::CreateRandomId();
    st >> id_str;

    item->AddAttr(buzz::QN_JID, jid.Str());
    item->AddAttr(buzz::QN_NAME, name);
    item_group->AddText(group);
    if (item_group->BodyText() != "")
        item->AddElement(item_group);

    query->AddAttr(buzz::QN_XMLNS, buzz::Constants::ns_roster());
    query->AddElement(item);

    result->AddAttr(buzz::QN_FROM, GetClient()->jid().Str());
    result->AddAttr(buzz::QN_TYPE, buzz::STR_SET);
    result->AddAttr(buzz::QN_ID, id_str);
    result->AddElement(query);

    subscribe->AddAttr(buzz::QN_TO, jid.Str());
    subscribe->AddAttr(buzz::QN_TYPE, buzz::STR_SUBSCRIBE);

//    SendStanza(result);
    SendStanza(subscribe);
}

void XmppRoster::RemoveContact(const buzz::Jid &jid)
{
    buzz::XmlElement *result = new buzz::XmlElement(buzz::QN_IQ);
    buzz::XmlElement *query = new buzz::XmlElement(buzz::QN_ROSTER_QUERY);
    buzz::XmlElement *item = new buzz::XmlElement(buzz::QN_ROSTER_ITEM);
    buzz::XmppReturnStatus status;
    DenyContact(jid);
    std::string id_str;
    std::stringstream st;
    st << cricket::CreateRandomId();
    st >> id_str;
    
    item->AddAttr(buzz::QN_JID, jid.Str());
    item->AddAttr(buzz::QN_SUBSCRIPTION, buzz::STR_REMOVE);

    query->AddAttr(buzz::QN_XMLNS, buzz::Constants::ns_roster());
    query->AddElement(item);

    result->AddAttr(buzz::QN_FROM, GetClient()->jid().Str());
    result->AddAttr(buzz::QN_TYPE, buzz::STR_SET);
    result->AddAttr(buzz::QN_ID, id_str);
    result->AddElement(query);
        
    status = SendStanza(result);
}

void XmppRoster::AuthorizeContact(const buzz::Jid &jid)
{
    buzz::XmlElement *result = new buzz::XmlElement(buzz::QN_PRESENCE);
    buzz::XmppReturnStatus status;

    result->AddAttr(buzz::QN_TO, jid.Str());
    result->AddAttr(buzz::QN_TYPE, buzz::STR_SUBSCRIBED);
    
    status = SendStanza(result);    
}

void XmppRoster::DenyContact(const buzz::Jid &jid)
{
    buzz::XmlElement *result = new buzz::XmlElement(buzz::QN_PRESENCE);
    buzz::XmppReturnStatus status;

    result->AddAttr(buzz::QN_TO, jid.Str());
    result->AddAttr(buzz::QN_TYPE, buzz::STR_UNSUBSCRIBED);
    
    status = SendStanza(result);
}

void XmppRoster::RequestContacts()
{
    buzz::XmlElement *result = new buzz::XmlElement(buzz::QN_IQ);
    buzz::XmlElement *query = new buzz::XmlElement(buzz::QN_ROSTER_QUERY);
    buzz::XmppReturnStatus status;
    std::string id_str;
    std::stringstream st;
    st << cricket::CreateRandomId();
    st >> id_str;
    
    query->AddAttr(buzz::QN_XMLNS, buzz::Constants::ns_roster());

    result->AddAttr(buzz::QN_FROM, GetClient()->jid().Str());
    result->AddAttr(buzz::QN_TYPE, buzz::STR_GET);
    result->AddAttr(buzz::QN_ID, id_str);
    result->AddElement(query);
        
    status = SendStanza(result);
}

void XmppRoster::RequestInfo(const buzz::Jid &jid)
{
    buzz::XmlElement *result = new buzz::XmlElement(buzz::QN_IQ);
}

bool XmppRoster::HandleStanza(const buzz::XmlElement *stanza) 
{
    /* Only handle Presence stanzas with types for subscriptions */
    if (stanza->Name() == buzz::QN_PRESENCE) {
        if (stanza->Attr(buzz::QN_TYPE) != buzz::STR_SUBSCRIBED && 
            stanza->Attr(buzz::QN_TYPE) != buzz::STR_SUBSCRIBE &&
            stanza->Attr(buzz::QN_TYPE) != buzz::STR_UNSUBSCRIBED)
            return false;
    }
    /* Only handle IQ stanzas with result type and <query xmlns='jabber:iq:roster'/>*/
    else if (stanza->Name() == buzz::QN_IQ) {
        if ((stanza->Attr(buzz::QN_TYPE) != buzz::STR_RESULT) ||
            stanza->FirstElement() == NULL ||
            stanza->FirstElement()->Name() != buzz::QN_ROSTER_QUERY)
            return false;
    }
    else
        return false;
    QueueStanza(stanza);
    return true;
}

int XmppRoster::ProcessStart() 
{
    const buzz::XmlElement *stanza = NextStanza();
    const buzz::XmlElement *query;
    const buzz::XmlElement *item;
    
    
    if (stanza == NULL)
        return STATE_BLOCKED;

    if (stanza->Name() == buzz::QN_IQ) {
        crit_.Enter();
        query = stanza->FirstNamed(buzz::QN_ROSTER_QUERY);
        for (item = query->FirstNamed(buzz::QN_ROSTER_ITEM); item;
            item = item->NextNamed(buzz::QN_ROSTER_ITEM)) {
            const buzz::XmlElement *group = item->FirstNamed(buzz::QN_ROSTER_GROUP);
            const std::string subscription = item->Attr(buzz::QN_SUBSCRIPTION);
            if (contact_cb_ && (subscription != buzz::STR_REMOVE) &&
                (item->Attr(buzz::QN_ASK) != buzz::STR_SUBSCRIBE)) {
                contact_cb_((xmpp_roster_t) this, contact_data_,
                    item->Attr(buzz::QN_JID).c_str(),
                    item->Attr(buzz::QN_NAME).c_str(),
                    group ? group->BodyText().c_str() : "",
                    subscription.c_str());
            }
        }
        crit_.Leave();
    }
    else if (stanza->Name() == buzz::QN_PRESENCE) {
        if ((stanza->Attr(buzz::QN_TYPE) == buzz::STR_SUBSCRIBED) && authorize_cb_)
            authorize_cb_((xmpp_roster_t) this, authorize_data_,
                stanza->Attr(buzz::QN_FROM).c_str());
        else if ((stanza->Attr(buzz::QN_TYPE) == buzz::STR_UNSUBSCRIBED) && deny_cb_)
            deny_cb_((xmpp_roster_t) this, deny_data_, stanza->Attr(buzz::QN_FROM).c_str());
        else if ((stanza->Attr(buzz::QN_TYPE) == buzz::STR_SUBSCRIBE) && request_cb_)
            request_cb_((xmpp_roster_t) this, request_data_,
                stanza->Attr(buzz::QN_FROM).c_str());
    }

    return STATE_START;
}
