/*
 * Tapioca XMPP Plugin
 * Copyright (c) 2005 INdT.
 * @author Andre Moreira Magalhaes <andre.magalhaes@indt.org.br> 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif 

#include "xmpp_text_session.h"

#include "xmpp_chat_c.h"
#include "xmpp_presence_push_c.h"

#include <tapioca/core/tapioca_enums.h>
#include <tapioca/core/tapioca_errors.h>
#include <tapioca/core/tapioca_utils.h>
#include <tapioca/core/tapioca_version.h>

#include <string.h>

struct _XmppTextSessionPrivate {
    xmpp_client_t client;
    xmpp_chat_t chat;
    gboolean disposed;
};

static void     xmpp_text_session_dispose     (GObject *object);

static gboolean xmpp_text_session_initiate    (TapiocaSession *ts,
                                               GError **error);
static gboolean xmpp_text_session_add_members (TapiocaSession *ts,
                                               const gchar **uris,
                                               GError **error);
static void     xmpp_text_session_close       (TapiocaSession *ts);
static gboolean xmpp_text_session_send        (TapiocaTextSession *tts,
                                               guint message_type,
                                               const gchar *subject,
                                               const gchar *body,
                                               GError **error);

static void     established_cb                (xmpp_chat_t chat, 
                                               gpointer data);
static void     message_received_cb           (xmpp_chat_t chat, 
                                               gpointer data,
                                               xmpp_message_t message);
static void     delivery_error_cb             (xmpp_chat_t chat, 
                                               gpointer data,
                                               xmpp_message_t message,
                                               int error_code);

G_DEFINE_TYPE(XmppTextSession, xmpp_text_session, TAPIOCA_TYPE_TEXT_SESSION)

static void 
xmpp_text_session_class_init (XmppTextSessionClass *klass)
{
    GObjectClass *gobject_class;
    TapiocaSessionClass *tapioca_session_class;
    TapiocaTextSessionClass *tapioca_text_session_class;

    gobject_class = (GObjectClass *) klass;
    tapioca_session_class = (TapiocaSessionClass*) klass;
    tapioca_text_session_class = (TapiocaTextSessionClass*) klass;
    xmpp_text_session_parent_class = g_type_class_peek_parent (klass);
    
    g_type_class_add_private (klass, sizeof (XmppTextSessionPrivate));

    tapioca_session_class->initiate = xmpp_text_session_initiate;
    tapioca_session_class->add_members = xmpp_text_session_add_members;
    tapioca_session_class->close = xmpp_text_session_close;
    
    tapioca_text_session_class->send = xmpp_text_session_send;

    gobject_class->dispose = xmpp_text_session_dispose;
}

static void 
xmpp_text_session_init (XmppTextSession *sc)
{
    sc->priv = XMPP_TEXT_SESSION_GET_PRIVATE (sc);
    sc->priv->disposed = FALSE;
}

static void
xmpp_text_session_dispose (GObject *object)
{
    XmppTextSession *sc = XMPP_TEXT_SESSION (object);

    if (sc->priv->disposed) {
        /* If dispose did already run, return. */
        return;
    }
    
    /* Make sure dispose does not run twice. */
    sc->priv->disposed = TRUE;
   
    G_OBJECT_CLASS (xmpp_text_session_parent_class)->dispose (object);
}

XmppTextSession *
xmpp_text_session_new (xmpp_client_t client)
{
    XmppTextSession *xts = XMPP_TEXT_SESSION (g_object_new (XMPP_TYPE_TEXT_SESSION, NULL));
    
    g_debug ("creating text session...");
    xts->priv->client = client;
    g_debug ("session created...");

    return xts;
}

XmppTextSession *
xmpp_text_session_new_with_chat (xmpp_client_t client,
                                 xmpp_chat_t chat)
{
    XmppTextSession *xts = XMPP_TEXT_SESSION (g_object_new (XMPP_TYPE_TEXT_SESSION, NULL));
    
    g_debug ("creating text session...");
    xts->priv->client = client;
    xts->priv->chat = chat;
    xmpp_chat_set_message_received_cb (chat, message_received_cb, xts);
    xmpp_chat_set_delivery_error_cb (chat, delivery_error_cb, xts);
    g_debug ("session created...");
    
    g_debug ("initiating chat");
    tapioca_session_change_state (TAPIOCA_SESSION (xts), TAPIOCA_SESSION_STATE_INITIATING);
    tapioca_session_change_state (TAPIOCA_SESSION (xts), TAPIOCA_SESSION_STATE_ESTABLISHED);
    g_debug ("chat in progress");

    return xts;
}

static gboolean
xmpp_text_session_initiate (TapiocaSession *ts,
                            GError **error)
{
    XmppTextSession *xts = XMPP_TEXT_SESSION (ts);
 
    g_set_error_val_if_fail (xts->priv->chat == NULL, FALSE, error, 
            TAPIOCA_ERROR, TAPIOCA_ERROR_GENERIC, "Cannot call initiate on an incoming chat");
   
    // create chat and initiate
    xts->priv->chat = xmpp_chat_session_client_create_chat (
            xmpp_client_get_chat_session_client (xts->priv->client), ts->members->data);
    xmpp_chat_set_message_received_cb (xts->priv->chat, message_received_cb, xts);
    xmpp_chat_set_delivery_error_cb (xts->priv->chat, delivery_error_cb, xts);

    /* session initiated with member bare jid
     * update member when session is established
     */
    if (!strchr(ts->members->data, '/'))
        xmpp_chat_set_established_cb (xts->priv->chat, established_cb, xts);

    g_debug ("initiating chat");
    tapioca_session_change_state (ts, TAPIOCA_SESSION_STATE_INITIATING);
    tapioca_session_change_state (ts, TAPIOCA_SESSION_STATE_ESTABLISHED);
    g_debug ("chat in progress");

    return TRUE;
}

static gboolean
xmpp_text_session_add_members (TapiocaSession *ts,
                               const gchar **jids,
                               GError **error)
{
    guint i;
    if (g_list_length (ts->members) > 0) {
        for (i = 0; jids[i]; ++i) 
            tapioca_session_signal_add_member_error (ts, TAPIOCA_ERROR_INVALID_PARAMETERS, jids[i]);
        g_set_error (error, TAPIOCA_ERROR, TAPIOCA_ERROR_INVALID_PARAMETERS, 
                "Cannot add more than one member to this session");
        return FALSE;
    }
    
    tapioca_session_signal_member_added (ts, jids[0]);
    for (i = 1; jids[i]; ++i) 
        tapioca_session_signal_add_member_error (ts, TAPIOCA_ERROR_INVALID_PARAMETERS, jids[i]);
    
    return TRUE;
}

static void
xmpp_text_session_close (TapiocaSession *ts)
{
    XmppTextSession *xts = XMPP_TEXT_SESSION (ts);
    xmpp_chat_session_client_destroy_chat (
            xmpp_client_get_chat_session_client (xts->priv->client),
            xts->priv->chat);
}

static gboolean 
xmpp_text_session_send (TapiocaTextSession *tts,
                        guint message_type,
                        const gchar *subject,
                        const gchar *body,
                        GError **error)
{
    XmppTextSession *xts = XMPP_TEXT_SESSION (tts);
       
    if (message_type == TAPIOCA_TEXT_SESSION_MESSAGE_TYPE_PLAIN_TEXT) {
        xmpp_chat_send_message (xts->priv->chat, subject, body);
    }
    else {
        g_set_error (error, TAPIOCA_ERROR, TAPIOCA_ERROR_NOT_SUPPORTED, "Message type not supported");
        return FALSE;
    }
    
    return TRUE;
}

/* this may happen when a chat was initialized with the member bare jid
 * and after the chat is established the member full jid should be used
 */
static void 
established_cb (xmpp_chat_t chat,
                gpointer data)
{
    TapiocaSession *ts = TAPIOCA_SESSION (data);    
    const gchar *old_member = (const gchar *) ts->members->data;
    const gchar *remote_jid = xmpp_chat_get_remote_jid (chat);
    
    tapioca_session_signal_member_added (ts, remote_jid);
    tapioca_session_signal_member_removed (ts, old_member);
}

static void 
message_received_cb (xmpp_chat_t chat, 
                     gpointer data,
                     xmpp_message_t message)
{
    TapiocaTextSession *tts = TAPIOCA_TEXT_SESSION (data);

    g_debug ("data=%p message received from %s - subject=%s body='%s'", 
            data,
            xmpp_message_get_from (message),
            xmpp_message_get_subject (message),
            xmpp_message_get_body (message));
    tapioca_text_session_signal_incoming (tts,
            xmpp_message_get_from (message),
            TAPIOCA_TEXT_SESSION_MESSAGE_TYPE_PLAIN_TEXT,
            xmpp_message_get_subject (message),
            xmpp_message_get_body (message));
}

static void 
delivery_error_cb (xmpp_chat_t chat, 
                   gpointer data,
                   xmpp_message_t message,
                   int error_code)
{
    TapiocaTextSession *tts = TAPIOCA_TEXT_SESSION (data);
    g_debug ("data=%p error sending message to %s - subject='%s' body='%s' error_code=%d", 
            data,
            xmpp_message_get_from (message),
            xmpp_message_get_subject (message), 
            xmpp_message_get_body (message), 
            error_code);
    tapioca_text_session_signal_delivery_error (tts,
            xmpp_message_get_from (message),
            TAPIOCA_TEXT_SESSION_MESSAGE_TYPE_PLAIN_TEXT,
            xmpp_message_get_subject (message), 
            xmpp_message_get_body (message),
            error_code);
}
